# Copyright 2023 The JAX Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implementation of Sparse Flash Attention, a.k.a. "Splash" attention."""

from __future__ import annotations

import dataclasses
import enum
import functools
from typing import Any, Callable, Literal, NamedTuple, Union, Optional, overload

import jax
from jax import ad_checkpoint
from jax import lax
from jax import tree_util
from jax.experimental import pallas as pl
from jax.experimental.pallas import tpu as pltpu
from . import splash_attention_mask as mask_lib
from . import splash_attention_mask_info as mask_info_lib
import jax.numpy as jnp
import numpy as np

partial = functools.partial
DEFAULT_MASK_VALUE = -0.7 * float(np.finfo(np.dtype("float32")).max)
NUM_LANES = 128
NUM_SUBLANES = 8
# We predefine some useful dimension numbers for dot_general
NN_DIM_NUMBERS = (((1,), (0,)), ((), ()))  # standard matmul
NT_DIM_NUMBERS = (((1,), (1,)), ((), ()))  # RHS transposed


# mypy: ignore-errors

class SegmentIds(NamedTuple):
    """SegmentIds for Q and KV sequences.

    SegmentIds are a mechanims to ensure that there is no cross-attention between
    segments (fraction of a sequence) that have been concatenated together into a
    sequence. Each array is a list of ids (integers). Only tokens with the same
    id are allowed to attend to each other.

    The static mask (e.g. causal) is "and-ed" with the segment id mask to form
    the actual attention mask. It is important that the latter does not have any
    all-zero rows (along dimension kv). Otherwise it would result in a invalid
    softmax (the denominator would be 0).
    This condition holds for causal self-attention because in this case segment
    ids form a block diagonal matrix so at least one element in each row is set.
    It is easy to break this condition with non-self-attention configurations.
    Attributes:
      q: segment ids along the Q sequence
      kv: segment ids along the KV sequence
    """

    q: jax.Array  # [q_seq_len]
    kv: jax.Array  # [kv_seq_len]


# Return type of SplashAttention function that implements the custom vjp rule.
SplashCustomReturnType = Union[
    # out, no residuals
    jax.Array,
        # out, residuals:
    tuple[jax.Array, tuple[jax.Array,]]
]

SplashResidualsType = tuple[
    jax.Array,  # q
    jax.Array,  # k
    jax.Array,  # v
    Optional[SegmentIds],  # segment_ids
    jax.Array,  # out
    jax.Array,  # logsumexp
    Optional[mask_info_lib.MaskInfo],  # dq_mask_info
    Optional[mask_info_lib.MaskInfo],  # dkv_mask_info
]

MaskFunctionType = Callable[..., jax.Array]


def get_kernel_name(
        is_mqa: bool, save_residuals: bool, is_segmented: bool, phase: str
) -> str:
    """Returns a unique name for all SplashAttention kernel variants."""

    assert phase == "dq" or phase == "dkv" or phase == "fwd"
    # Saving residuals is supported only for the fwd phase.
    assert not save_residuals or phase == "fwd"
    residuals = ""
    if save_residuals:
        residuals = "_residuals"
    elif phase == "fwd":
        residuals = "_no_residuals"
    attention_type = "mqa" if is_mqa else "mha"
    segments = "_segmented" if is_segmented else ""
    return f"splash_{attention_type}_{phase}{segments}{residuals}"


# Reference attention implementations


@overload
def _attention_reference(
        mask: jax.Array,
        q: jax.Array,
        k: jax.Array,
        v: jax.Array,
        segment_ids: Optional[SegmentIds],
        save_residuals: Literal[False],
        mask_value: float,
        custom_type: str,
        attn_logits_soft_cap: Optional[float],
) -> jax.Array:
    ...


@overload
def _attention_reference(
        mask: jax.Array,
        q: jax.Array,
        k: jax.Array,
        v: jax.Array,
        segment_ids: Optional[SegmentIds],
        save_residuals: Literal[True],
        mask_value: float,
        custom_type: str,
        attn_logits_soft_cap: Optional[float],
) -> tuple[jax.Array, tuple[jax.Array]]:
    ...


def _attention_reference(
        mask: jax.Array,  # [q_seq_len, kv_seq_len]
        q: jax.Array,  # [q_seq_len, head_dim]
        k: jax.Array,  # [kv_seq_len, head_dim]
        v: jax.Array,  # [kv_seq_len, head_dim]
        segment_ids: Optional[SegmentIds],
        mask_value: float,
        save_residuals: bool,
        custom_type: str,
        attn_logits_soft_cap: Optional[float],
):
    return _attention_reference_default(  # pytype: disable=bad-return-type
        mask,
        q,
        k,
        v,
        segment_ids,
        mask_value,
        save_residuals,
        custom_type,
        attn_logits_soft_cap,
    )


def _attention_reference_default(
        mask: jax.Array,  # [q_seq_len, kv_seq_len]
        q: jax.Array,  # [q_seq_len, head_dim]
        k: jax.Array,  # [kv_seq_len, head_dim]
        v: jax.Array,  # [kv_seq_len, head_dim]
        segment_ids: Optional[SegmentIds],
        mask_value: float,
        save_residuals: bool,
        custom_type: str,
        attn_logits_soft_cap: Optional[float],
):
    del custom_type
    logits = jnp.einsum("sd,td->st", q.astype(jnp.float32), k.astype(jnp.float32))

    if segment_ids is not None:
        mask = jnp.logical_and(
            mask, segment_ids.q[:, None] == segment_ids.kv[None, :]
        )

    if attn_logits_soft_cap is not None:
        logits = jnp.tanh(logits / attn_logits_soft_cap)
        logits = logits * attn_logits_soft_cap

    logits = jnp.where(mask, logits, mask_value)
    m = logits.max(axis=-1)
    s = jnp.exp(logits - m[..., None])
    l = s.sum(axis=-1)
    s = s / l[..., None]

    o = jnp.einsum("st,td->sd", s, v.astype(jnp.float32))

    logsumexp = m + jnp.log(l)
    if save_residuals:
        return o, (logsumexp,)
    return o


def attention_reference(
        mask: jax.Array,  # [q_seq_len, kv_seq_len]
        q: jax.Array,  # [q_seq_len, head_dim]
        k: jax.Array,  # [kv_seq_len, head_dim]
        v: jax.Array,  # [kv_seq_len, head_dim]
        segment_ids: Optional[SegmentIds],
        *,
        mask_value: float = DEFAULT_MASK_VALUE,
        save_residuals: bool = False,
        custom_type: str = "flash",
        attn_logits_soft_cap: Optional[float] = None,
) -> SplashCustomReturnType:
    return _attention_reference(  # pytype: disable=wrong-arg-types
        mask,
        q,
        k,
        v,
        segment_ids,
        mask_value=mask_value,
        save_residuals=save_residuals,
        custom_type=custom_type,
        attn_logits_soft_cap=attn_logits_soft_cap,
    )


def _attention_reference_custom_fwd(
        mask: jax.Array,  # [q_seq_len, kv_seq_len]
        q: jax.Array,  # [q_seq_len, head_dim]
        k: jax.Array,  # [kv_seq_len, head_dim]
        v: jax.Array,  # [kv_seq_len, head_dim]
        segment_ids: Optional[SegmentIds],
        mask_value: float,
        save_residuals: bool,
        custom_type: str,
        attn_logits_soft_cap: Optional[float],
):
    if save_residuals:
        raise NotImplementedError("Higher-order AD not supported.")

    o, (logsumexp,) = _attention_reference(
        mask,
        q,
        k,
        v,
        segment_ids,
        mask_value=mask_value,
        save_residuals=True,
        custom_type=custom_type,
        attn_logits_soft_cap=attn_logits_soft_cap,
    )
    return o, (mask, q, k, v, segment_ids, o, logsumexp)


def _attention_reference_custom_bwd(
        mask_value: float,
        save_residuals: bool,
        custom_type: str,
        attn_logits_soft_cap: Optional[float],
        res,
        do: jax.Array,
) -> tuple[None, jax.Array, jax.Array, jax.Array, None]:
    del save_residuals
    mask, q, k, v, segment_ids, o, logsumexp = res

    uncapped_logits = jnp.einsum(
        "qc,kc->qk", q, k, preferred_element_type=jnp.float32)

    if attn_logits_soft_cap is not None:
        logits = jnp.tanh(uncapped_logits / attn_logits_soft_cap)
        logits = logits * attn_logits_soft_cap
    else:
        logits = uncapped_logits

    if segment_ids is not None:
        mask = jnp.logical_and(
            mask, segment_ids.q[:, None] == segment_ids.kv[None, :]
        )
    logits = jnp.where(mask, logits, mask_value)

    p = jnp.exp(logits - logsumexp[..., None])
    do = do.astype(jnp.float32)  # pytype: disable=attribute-error
    dv = jnp.einsum("pt,pd->td", p, do).astype(v.dtype)
    dp = jnp.einsum("pd,td->pt", do, v.astype(jnp.float32))

    # These two ways of computing ds are mathematically equivalent. The first
    # involves reducing over the head_dim dimension and the second involves
    # reducing over a sequence dimension. They tend to produce slightly different
    # numerics.
    if custom_type == "flash":
        di = jnp.sum(o.astype(jnp.float32) * do, axis=-1)[..., None]
    else:
        di = jnp.einsum("st,st->s", dp, p)[:, None]
    ds = (dp - di) * p
    if attn_logits_soft_cap is not None:
        normalized = uncapped_logits / attn_logits_soft_cap
        d = jnp.tanh(normalized)
        g = ds * (1 - d)
        ds = g + g * d
    dk = jnp.einsum("sd,st->td", q.astype(jnp.float32), ds).astype(k.dtype)
    dq = jnp.einsum("st,td->sd", ds, k.astype(jnp.float32)).astype(q.dtype)
    return None, dq, dk, dv, None


_attention_reference_custom = jax.custom_vjp(
    _attention_reference, nondiff_argnums=(5, 6, 7, 8)
)
_attention_reference_custom.defvjp(_attention_reference_custom_fwd,
                                   _attention_reference_custom_bwd)


def attention_reference_custom(
        mask: jax.Array,  # [q_seq_len, kv_seq_len]
        q: jax.Array,  # [q_seq_len, head_dim]
        k: jax.Array,  # [kv_seq_len, head_dim]
        v: jax.Array,  # [kv_seq_len, head_dim]
        segment_ids: Optional[SegmentIds],
        *,
        mask_value: float = DEFAULT_MASK_VALUE,
        save_residuals: bool = False,
        custom_type: str = "flash",
        attn_logits_soft_cap: Optional[float] = None,
):
    return _attention_reference_custom(
        mask,
        q,
        k,
        v,
        segment_ids,
        mask_value,
        save_residuals,
        custom_type=custom_type,
        attn_logits_soft_cap=attn_logits_soft_cap,
    )


def make_attention_reference(
        mask: Union[mask_lib.Mask , np.ndarray],
        is_mqa: bool,
        backward_impl: str = "vanilla",
        **params: Any,
) -> Callable:
    @partial(
        jax.jit,
        static_argnames=[
            "mask_value",
            "save_residuals",
            "attn_logits_soft_cap",
        ],
    )
    def _wrapped(
            mask: jax.Array,
            q: jax.Array,
            k: jax.Array,
            v: jax.Array,
            segment_ids: Optional[SegmentIds] = None,
            *,
            mask_value: float = DEFAULT_MASK_VALUE,
            save_residuals: bool = False,
            attn_logits_soft_cap: Optional[float] = None,
    ):
        if backward_impl == "custom":
            attn_impl = partial(
                attention_reference_custom, custom_type="flash",
            )
        elif backward_impl == "custom_vanilla":
            attn_impl = partial(
                attention_reference_custom, custom_type="vanilla",
            )
        else:
            attn_impl = attention_reference
        func = partial(
            attn_impl,
            mask_value=mask_value,
            save_residuals=save_residuals,
            attn_logits_soft_cap=attn_logits_soft_cap,
            **params,
        )

        if is_mqa:
            func = jax.vmap(func, in_axes=(0, 0, None, None, None))
            is_grouped = False
        else:
            # In grouped attention (1 < num_kv_heads && num_kv_heads < num_q_heads).
            # We interleave the KV heads across the Q heads.
            # For example: for 8 Q heads and 4 KV heads:
            # Q head [0, 1] see KV head 0
            # Q head [2, 3] see KV head 1
            # Q head [4, 5] see KV head 2
            # Q head [6, 7] see KV head 3
            #
            # The following implementation reshapes Q to expose KV heads and vmaps
            # Across the Q heads so it is similar to MQA.
            # Alternatively we can replicate K/V to match Q like so:
            # k = jnp.repeat(k, q_heads_per_kv_head, axis=0)
            # v = jnp.repeat(v, q_heads_per_kv_head, axis=0)

            kv_heads = k.shape[0]
            assert kv_heads == v.shape[0]
            q_heads, q_seq_len, head_dim = q.shape
            is_grouped = kv_heads < q_heads
            if is_grouped:
                assert q_heads % kv_heads == 0
                assert mask.shape[0] == q_heads
                q_heads_per_kv_head = q_heads // kv_heads
                q = q.reshape((kv_heads, q_heads_per_kv_head, q_seq_len, head_dim))
                mask = mask.reshape((kv_heads, q_heads_per_kv_head, *mask.shape[1:]))

                # Inner-most vmap: iterate over the q heads.
                func = jax.vmap(func, in_axes=(0, 0, None, None, None))

            # Outer-most vmap: iterate over the kv heads.
            func = jax.vmap(func, in_axes=(0, 0, 0, 0, None))

        out = func(mask, q, k, v, segment_ids)

        if is_grouped:

            def reshape_activations(activations):
                if activations.ndim == 4:
                    kv_heads, q_heads_per_kv_head, q_seq_len, head_dim = activations.shape
                    return activations.reshape(
                        kv_heads * q_heads_per_kv_head, q_seq_len, head_dim
                    )
                return activations

            def reshape_residuals(residuals):
                if residuals.ndim == 3:
                    kv_heads, q_heads_per_kv_head, q_seq_len = residuals.shape
                    return residuals.reshape(kv_heads * q_heads_per_kv_head, q_seq_len)
                return residuals

            if save_residuals:
                assert isinstance(out, tuple)
                assert isinstance(out[1], tuple)

                return (reshape_activations(out[0]), (reshape_residuals(out[1][0]),))
            else:
                return reshape_activations(out)
        else:
            return out

    return functools.partial(_wrapped, jnp.array(mask[:, :, :]))


make_masked_mha_reference = partial(make_attention_reference, is_mqa=False)
make_masked_mqa_reference = partial(make_attention_reference, is_mqa=True)


# Splash attention implementation

# We use an IntEnum to make it JSON serializable as regen metadata.
class QKVLayout(enum.IntEnum):
    HEAD_DIM_MINOR = enum.auto()  # [..., seq_len, head_dim]
    SEQ_MINOR = enum.auto()  # [..., head_dim, seq_len]


def from_head_minor(vals: tuple[Any, ...], layout: QKVLayout):
    if layout == QKVLayout.HEAD_DIM_MINOR:
        return vals
    return (*vals[:-2], vals[-1], vals[-2])


@dataclasses.dataclass(unsafe_hash=True)
class BlockSizes:
    """Tile sizes parameterizing SplashAttention kernels.

    Those parameters have negligible effect on numerics, but affect performance
    greatly.

    Note that changing the layouts only influences the physical layout that the
    kernel will enforce. The logical interface to splash attention always takes
    the head dimension as the minormost one.
    """
    block_q: int
    block_kv: int
    block_kv_compute: Optional[int] = None

    block_q_dkv: Optional[int] = None
    block_kv_dkv: Optional[int] = None
    block_kv_dkv_compute: Optional[int] = None

    block_q_dq: Optional[int] = None
    block_kv_dq: Optional[int] = None

    use_fused_bwd_kernel: bool = False

    q_layout: QKVLayout = QKVLayout.HEAD_DIM_MINOR
    k_layout: QKVLayout = QKVLayout.HEAD_DIM_MINOR
    v_layout: QKVLayout = QKVLayout.HEAD_DIM_MINOR

    def __post_init__(self):
        if self.block_kv_compute is None:
            self.block_kv_compute = self.block_kv
        if self.block_kv_dkv_compute is None:
            self.block_kv_dkv_compute = self.block_kv_dkv
        if self.use_fused_bwd_kernel:
            if self.block_q_dq is not None or self.block_kv_dq is not None:
                raise ValueError(
                    "Block sizes for dq kernel are not needed with a fused kernel."
                )

    @property
    def has_backward_blocks(self) -> bool:
        backward_blocks = (
            self.block_q_dkv, self.block_kv_dkv, self.block_kv_dkv_compute,
        )
        if not self.use_fused_bwd_kernel:
            backward_blocks += (self.block_q_dq, self.block_kv_dq)
        return all(b is not None for b in backward_blocks)

    @classmethod
    def get_default(cls):
        # TODO(apaszke,sharadmv): Select better parameters based on a heuristic.
        return BlockSizes(
            block_q=128,
            block_kv=128,
            block_kv_compute=128,
            block_q_dkv=128,
            block_kv_dkv=128,
            block_kv_dkv_compute=128,
            block_q_dq=128,
            block_kv_dq=128,
        )


def _next_nonzero(
        h,
        i,
        j,
        data_next_ref,
        block_mask_ref,
        m_next_ref,
        next_i=False,
):
    assert (data_next_ref is None) == (block_mask_ref is None)

    if data_next_ref is None and block_mask_ref is None:
        # Handle the case in which we have no masking nor next data information.
        # Simply fetch the next data and apply the mask for every block.
        assert m_next_ref is None
        next_data = i if next_i else j
        return (
            next_data,
            None,  # next mask
            True,  # should run
            False,  # should not mask
        )

    assert data_next_ref.shape == block_mask_ref.shape
    assert m_next_ref is None or data_next_ref.shape[0] == m_next_ref.shape[0]

    # We are working with one head only. Force the head index to 0.
    if data_next_ref.shape[0] == 1:
        h = 0

    # When scalar-memory data is of types smaller than int32, then we have to
    # upcast it back to use it in the kernel.

    to_i32 = lambda x: x.astype(jnp.int32)

    is_nonzero = to_i32(block_mask_ref[h, i, j]) > 0
    if m_next_ref is None:
        should_not_mask = True
        next_m = None
    else:
        should_not_mask = to_i32(block_mask_ref[h, i, j]) != 1
        next_m = to_i32(m_next_ref[h, i, j])
    next_j = to_i32(data_next_ref[h, i, j])
    return next_j, next_m, is_nonzero, should_not_mask


def _apply_mask_and_soft_cap(
        qk: jax.Array,
        mask_value: float,
        should_not_mask,
        mask_ref,
        q_sequence_ref,
        q_segment_ids_ref,
        kv_segment_ids_ref,
        *,
        attn_logits_soft_cap: float,
        k_slice: pl.Slice,
        k_offset: int,
        bq: int,
        k_in_lanes=True,
        mask_function=None,
) -> Union[jax.Array , tuple[jax.Array, jax.Array, jax.Array, jax.Array]]:
    assert mask_ref is None or q_sequence_ref is None
    assert (q_sequence_ref is None) == (mask_function is None)

    masks = []
    if mask_ref is not None:
        if k_in_lanes:
            mask = pl.load(mask_ref, (slice(None), k_slice))
        else:
            mask = pl.load(mask_ref, (k_slice, slice(None)))

        snm = jnp.where(should_not_mask, 1, 0)
        masks.append(jnp.bitwise_or(mask, jnp.broadcast_to(snm, mask.shape)))

    if mask_function is not None:
        # Compute the mask using the given q_sequence indices.
        # KV indices are computed on the fly. This works because we only support Q
        # sequence sharding. If we wanted to compute Q indices too, then we would
        # need to keep into account the current shard along Q sequence.

        if k_in_lanes:
            assert q_sequence_ref.shape == (bq, NUM_LANES)

            k_sequence = k_offset + jax.lax.broadcasted_iota(
                jnp.int32, (bq, k_slice.size), 1
            )

            repeats, rem = divmod(k_slice.size, NUM_LANES)
            assert rem == 0
            q_sequence = pltpu.repeat(
                q_sequence_ref[...], repeats, axis=1
            )  # [bq, k_slice.size]
        else:
            assert q_sequence_ref.shape == (NUM_SUBLANES, bq)

            k_sequence = k_offset + jax.lax.broadcasted_iota(
                jnp.int32, (k_slice.size, bq), 0
            )
            q_sequence = pl.load(q_sequence_ref, (pl.ds(1), slice(None)))  # [1, bq]
            q_sequence = jnp.broadcast_to(q_sequence, (k_slice.size, bq))

        assert q_sequence.shape == k_sequence.shape
        masks.append(mask_function(q_sequence, k_sequence))  # pytype: disable=wrong-arg-count

    if q_segment_ids_ref is not None:
        if k_in_lanes:
            kv_ids = pl.load(kv_segment_ids_ref, (pl.ds(1), k_slice))  # [1, k_slice]
            repeats, rem = divmod(kv_ids.shape[1], NUM_LANES)
            if rem:
                raise NotImplementedError(f"block_kv must be a multiple of {NUM_LANES}")
            q_ids = pltpu.repeat(q_segment_ids_ref[:], repeats, axis=1)  # [bq, bkv]
        else:
            assert bq == q_segment_ids_ref.shape[-1]
            repeats, rem = divmod(bq, NUM_LANES)
            if rem:
                raise NotImplementedError(f"block_q must be a multiple of {NUM_LANES}")
            kv_ids = pltpu.repeat(
                pl.load(kv_segment_ids_ref, (k_slice, slice(None))), repeats, axis=1
            )  # [k_slice, bq]
            q_ids = pl.load(q_segment_ids_ref, (pl.ds(1), slice(None)))  # [1, bq]
        masks.append(q_ids == kv_ids)

    def cap_logits(logits):
        if attn_logits_soft_cap is not None:
            logits = jnp.tanh(qk / attn_logits_soft_cap)
            return logits * attn_logits_soft_cap
        else:
            return logits

    if masks:
        mask = functools.reduce(jnp.logical_and, masks)
        qk = cap_logits(qk)
        qk = jnp.where(mask, qk, mask_value)
    else:
        qk = cap_logits(qk)
    return qk


def flash_attention_kernel(
        # Prefetched inputs
        data_next_ref,
        block_mask_ref,
        mask_next_ref,
        # Inputs
        q_ref,
        k_ref,
        v_ref,
        q_segment_ids_ref,
        kv_segment_ids_ref,
        mask_ref,
        q_sequence_ref,
        # Outputs
        m_scratch_ref,
        l_scratch_ref,
        o_scratch_ref,
        o_ref,
        logsumexp_ref=None,
        *,
        mask_value: float,
        grid_width: int,
        bq: int,
        bkv: int,
        bkv_compute: int,
        head_dim: int,
        q_layout: QKVLayout,
        k_layout: QKVLayout,
        v_layout: QKVLayout,
        attn_logits_soft_cap: Optional[float],
        mask_function: Optional[MaskFunctionType],
):
    float32 = jnp.float32
    HEAD_DIM_MINOR = QKVLayout.HEAD_DIM_MINOR

    head_dim_repeats, rem = divmod(head_dim, NUM_LANES)
    if rem != 0:
        raise NotImplementedError(
            f"{head_dim=} should be a multiple of {NUM_LANES}"
        )

    h, i, j = pl.program_id(0), pl.program_id(1), pl.program_id(2)

    @pl.when(j == 0)
    def init():
        o_scratch_ref[...] = jnp.zeros_like(o_scratch_ref)
        m_scratch_ref[...] = jnp.full_like(m_scratch_ref, mask_value)
        l_scratch_ref[...] = jnp.zeros_like(l_scratch_ref)

    _, _, should_run, should_not_mask = _next_nonzero(
        h,
        i,
        j,
        data_next_ref,
        block_mask_ref,
        mask_next_ref,
    )

    def body(kv_compute_index, _):
        slice_k = pl.ds(kv_compute_index * bkv_compute, bkv_compute)
        m_prev, l_prev = m_scratch_ref[...], l_scratch_ref[...]
        assert m_prev.shape == (bq, NUM_LANES)
        assert l_prev.shape == (bq, NUM_LANES)

        q = q_ref[...] if q_layout == HEAD_DIM_MINOR else q_ref[...].T
        qk_dims = NT_DIM_NUMBERS if k_layout == HEAD_DIM_MINOR else NN_DIM_NUMBERS
        if k_layout == HEAD_DIM_MINOR:
            k = pl.load(k_ref, (slice_k, slice(None)))
        else:
            k = pl.load(k_ref, (slice(None), slice_k))
        qk = lax.dot_general(q, k, qk_dims, preferred_element_type=float32)

        assert qk.shape == (bq, bkv_compute)
        apply_mask_and_soft_cap = functools.partial(
            _apply_mask_and_soft_cap,
            qk,
            mask_value,
            should_not_mask,
            mask_ref,
            q_sequence_ref,
            q_segment_ids_ref,
            kv_segment_ids_ref,
            attn_logits_soft_cap=attn_logits_soft_cap,
            k_slice=slice_k,
            k_offset=j * bkv + kv_compute_index * bkv_compute,
            bq=bq,
            mask_function=mask_function,
        )

        qk = apply_mask_and_soft_cap()

        m_curr = qk.max(axis=-1)[:, None]
        assert m_curr.shape == (bq, 1)
        m_next = jnp.maximum(m_prev, m_curr)
        assert m_next.shape == (bq, NUM_LANES)

        bkv_repeats, rem = divmod(bkv_compute, NUM_LANES)
        if rem != 0:
            raise NotImplementedError(
                f"{bkv_compute=} should be a multiple of {NUM_LANES}"
            )

        s_curr = jnp.exp(qk - pltpu.repeat(m_next, bkv_repeats, axis=1))
        assert s_curr.shape == (bq, bkv_compute)

        l_curr = jax.lax.broadcast_in_dim(s_curr.sum(axis=-1), l_prev.shape, (0,))
        assert l_curr.shape == (bq, NUM_LANES)

        alpha = jnp.exp(m_prev - m_next)
        l_next = l_curr + alpha * l_prev
        m_scratch_ref[...], l_scratch_ref[...] = m_next, l_next

        sv_dims = NN_DIM_NUMBERS if v_layout == HEAD_DIM_MINOR else NT_DIM_NUMBERS
        if v_layout == HEAD_DIM_MINOR:
            v = pl.load(v_ref, (slice_k, slice(None)))
        else:
            v = pl.load(v_ref, (slice(None), slice_k))
        v = v.astype(float32)
        o_curr = lax.dot_general(s_curr, v, sv_dims)

        alpha_o = pltpu.repeat(alpha, head_dim_repeats, axis=1)
        o_scratch_ref[:] = alpha_o * o_scratch_ref[:] + o_curr

    @pl.when(should_run)
    def run():
        assert bkv % bkv_compute == 0
        num_iters = (
                k_ref.shape[0 if k_layout == HEAD_DIM_MINOR else 1] // bkv_compute
        )
        lax.fori_loop(0, num_iters, body, None, unroll=True)

    @pl.when(j == grid_width - 1)
    def end():
        l = l_scratch_ref[...]
        l_inv = pltpu.repeat(1.0 / l, head_dim_repeats, axis=1)
        o_ref[...] = (o_scratch_ref[...] * l_inv).astype(o_ref.dtype)
        if logsumexp_ref is not None:
            assert logsumexp_ref.shape == (bq, NUM_LANES)
            logsumexp_ref[...] = (jnp.log(l) + m_scratch_ref[...]).astype(
                logsumexp_ref.dtype
            )

        m_scratch_ref[...] = jnp.zeros_like(m_scratch_ref)
        l_scratch_ref[...] = jnp.zeros_like(l_scratch_ref)
        o_scratch_ref[...] = jnp.zeros_like(o_scratch_ref)


@overload
def _splash_attention_forward(
        fwd_mask_info: mask_info_lib.MaskInfo,
        q: jax.Array,
        k: jax.Array,
        v: jax.Array,
        segment_ids: Optional[SegmentIds],
        mask_value: float,
        is_mqa: bool,
        block_sizes: BlockSizes,
        residual_checkpoint_name: Optional[str],
        mask_function: Optional[MaskFunctionType],
        save_residuals: Literal[False] = False,
        attn_logits_soft_cap: Optional[float] = None,
) -> jax.Array:
    ...


@overload
def _splash_attention_forward(
        fwd_mask_info: mask_info_lib.MaskInfo,
        q: jax.Array,
        k: jax.Array,
        v: jax.Array,
        segment_ids: Optional[SegmentIds],
        mask_value: float,
        is_mqa: bool,
        block_sizes: BlockSizes,
        residual_checkpoint_name: Optional[str],
        mask_function: Optional[MaskFunctionType],
        save_residuals: Literal[True],
        attn_logits_soft_cap: Optional[float] = None,
) -> SplashCustomReturnType:
    ...


def _div(dividend: int, divisor: int):
    if divisor == 1:
        return dividend

    return lax.div(dividend, divisor)


def _splash_attention_forward(
        fwd_mask_info: mask_info_lib.MaskInfo,
        q: jax.Array,
        k: jax.Array,
        v: jax.Array,
        segment_ids: Optional[SegmentIds],
        mask_value: float,
        is_mqa: bool,
        block_sizes: BlockSizes,
        residual_checkpoint_name: Optional[str],
        save_residuals: bool,
        mask_function: Optional[MaskFunctionType],
        attn_logits_soft_cap: Optional[float] = None,
) -> SplashCustomReturnType:
    num_q_heads, q_seq_len, head_dim = q.shape
    bq, bkv = block_sizes.block_q, block_sizes.block_kv
    bkv_compute = block_sizes.block_kv_compute

    if is_mqa:
        expected_kv_rank = 2
        kv_head_dimension = 1
        kv_seq_len_dimension = 0
        num_kv_heads = 1
    else:
        expected_kv_rank = 3
        kv_head_dimension = 2
        kv_seq_len_dimension = 1
        num_kv_heads = k.shape[0]

    if len(k.shape) != expected_kv_rank:
        raise ValueError(
            f"Expected {expected_kv_rank}-dim 'key' tensor for MQA. Instead got a"
            f" {len(k.shape)}-dim one."
        )

    if k.shape[kv_head_dimension] != head_dim:
        raise ValueError(
            f"Expected 'key' head dimension to be: {head_dim}. Instead got:"
            f" {k.shape[kv_head_dimension]}."
        )

    if not is_mqa and num_q_heads % num_kv_heads != 0:
        raise ValueError(
            f"In MHA, expected number of 'key' heads ({num_kv_heads}) to be a"
            f" multiple of the number of 'query' heads ({num_q_heads})"
        )

    if k.shape != v.shape:
        raise ValueError(
            f"Expected 'key' {k.shape} and 'value' {v.shape} to have the same"
            " shape."
        )

    if bkv % bkv_compute:
        raise ValueError(f"{bkv=} must be a multiple of {bkv_compute=}.")
    if bkv_compute % NUM_LANES:
        raise ValueError(f"{bkv=} must be a multiple of {NUM_LANES}.")

    kv_seq_len = k.shape[kv_seq_len_dimension]

    q_heads_per_kv_head = num_q_heads // num_kv_heads

    if segment_ids is not None:
        if segment_ids.q.shape != (q_seq_len,):
            raise ValueError(
                "Invalid shape for q segment_ids: "
                f"{segment_ids.q.shape}. Expected: {(q_seq_len,)}"
            )
        if segment_ids.kv.shape != (kv_seq_len,):
            raise ValueError(
                "Invalid shape for kv segment_ids: "
                f"{segment_ids.kv.shape}. Expected: {(kv_seq_len,)}"
            )

    q_layout = block_sizes.q_layout

    def q_index_map(h, i, j, data_next_ref, block_mask_ref, mask_next_ref=None):
        del j, data_next_ref, mask_next_ref, block_mask_ref
        return from_head_minor((h, i, 0), q_layout)

    def out_index_map(h, i, j, data_next_ref, block_mask_ref, mask_next_ref=None):
        del j, data_next_ref, mask_next_ref, block_mask_ref
        return h, i, 0

    k_layout = block_sizes.k_layout

    def k_index_map(h, i, j, data_next_ref, block_mask_ref, mask_next_ref=None):
        next_j, *_ = _next_nonzero(
            h, i, j, data_next_ref, block_mask_ref, mask_next_ref
        )
        prefix = () if is_mqa else (_div(h, q_heads_per_kv_head),)
        return from_head_minor((*prefix, next_j, 0), k_layout)

    v_layout = block_sizes.v_layout

    def v_index_map(h, i, j, data_next_ref, block_mask_ref, mask_next_ref=None):
        next_j, *_ = _next_nonzero(
            h, i, j, data_next_ref, block_mask_ref, mask_next_ref
        )
        prefix = () if is_mqa else (_div(h, q_heads_per_kv_head),)
        return from_head_minor((*prefix, next_j, 0), v_layout)

    def mask_index_map(h, i, j, data_next_ref, block_mask_ref,
                       mask_next_ref=None):
        _, next_m, *_ = _next_nonzero(
            h, i, j, data_next_ref, block_mask_ref, mask_next_ref
        )
        return next_m, 0, 0

    def q_segment_ids_index_map(h, i, j, *_):
        del h, j  # Unused.
        return i, 0

    def kv_segment_ids_index_map(h, i, j, data_next_ref, block_mask_ref,
                                 mask_next_ref=None):
        next_j, *_ = _next_nonzero(
            h, i, j, data_next_ref, block_mask_ref, mask_next_ref
        )
        return 0, next_j

    # Convert the logical shape from head-minor to sequence-minor.
    in_specs = [
        pl.BlockSpec(
            q_index_map, from_head_minor((None, bq, head_dim), q_layout)
        ),
        pl.BlockSpec(
            k_index_map,
            from_head_minor(
                (bkv, head_dim) if is_mqa else (None, bkv, head_dim), k_layout
            ),
        ),
        pl.BlockSpec(
            v_index_map,
            from_head_minor(
                (bkv, head_dim) if is_mqa else (None, bkv, head_dim), v_layout
            ),
        ),
    ]
    if segment_ids is not None:
        in_specs += [
            pl.BlockSpec(q_segment_ids_index_map, (bq, NUM_LANES)),
            pl.BlockSpec(kv_segment_ids_index_map, (NUM_SUBLANES, bkv)),
        ]
        q_segment_ids = jax.lax.broadcast_in_dim(
            segment_ids.q, (q_seq_len, NUM_LANES), (0,)
        )
        kv_segment_ids = jax.lax.broadcast_in_dim(
            segment_ids.kv, (NUM_SUBLANES, kv_seq_len), (1,)
        )
    else:
        in_specs += [None, None]
        q_segment_ids = kv_segment_ids = None

    if fwd_mask_info.partial_mask_blocks is not None:
        in_specs.append(pl.BlockSpec(mask_index_map, (None, bq, bkv)))
    else:
        in_specs.append(None)

    assert (
            fwd_mask_info.partial_mask_blocks is None
            or fwd_mask_info.q_sequence is None
    )

    if fwd_mask_info.q_sequence is not None:
        q_sequence = jax.lax.broadcast_in_dim(
            fwd_mask_info.q_sequence, (q_seq_len, NUM_LANES), (0,)
        )
        in_specs.append(pl.BlockSpec(q_segment_ids_index_map, (bq, NUM_LANES)))
    else:
        q_sequence = None
        in_specs.append(None)

    num_scalar_prefetch = 3

    out_shapes = [
        jax.ShapeDtypeStruct((bq, NUM_LANES), jnp.float32),  # m_scratch
        jax.ShapeDtypeStruct((bq, NUM_LANES), jnp.float32),  # l_scratch
        jax.ShapeDtypeStruct((bq, head_dim), jnp.float32),  # o_scratch
        jax.ShapeDtypeStruct((num_q_heads, q_seq_len, head_dim), q.dtype),
    ]
    out_specs = [
        # TODO(sharadmv): convert m/l to be scratch
        pl.BlockSpec(lambda h, i, j, *_: (0, 0), (bq, NUM_LANES)),
        pl.BlockSpec(lambda h, i, j, *_: (0, 0), (bq, NUM_LANES)),
        pl.BlockSpec(lambda h, i, j, *_: (0, 0), (bq, head_dim)),
        pl.BlockSpec(out_index_map, (None, bq, head_dim)),
    ]
    if save_residuals:
        out_shapes += [
            jax.ShapeDtypeStruct(
                (num_q_heads, q_seq_len, NUM_LANES), jnp.float32
            ),  # logsumexp
        ]

        def logsumexp_index_map(h, i, *_):
            return h, i, 0

        out_specs += [
            pl.BlockSpec(logsumexp_index_map, (None, bq, NUM_LANES)),
        ]
    else:
        out_shapes += [None]
        out_specs += [None]

    # Attach useful metadata to the custom-call HLO op.
    # Having this information available in an HLO-dump or xprof is valuable for
    # debugging and performance investigation.
    metadata_dict = dict(
        block_sizes=dataclasses.asdict(block_sizes),
        is_mqa=is_mqa,
        save_residuals=save_residuals,
        mask_value=mask_value,
        is_segmented=segment_ids is not None,
        attn_logits_soft_cap=attn_logits_soft_cap,
        residual_checkpoint_name=residual_checkpoint_name,
    )

    mosaic_params = pltpu.encode_kernel_regeneration_metadata(metadata_dict)

    mosaic_params.update(
        dimension_semantics=("parallel", "arbitrary", "arbitrary"),
        flags={"XLA_TPU_FORCE_LP_LLO_SCHEDULER": True},
    )

    kernel_name = get_kernel_name(
        is_mqa, save_residuals, segment_ids is not None, "fwd"
    )

    if fwd_mask_info.data_next is not None:
        grid_width = fwd_mask_info.data_next.shape[-1]
    else:
        grid_width = kv_seq_len // bkv

    grid = (num_q_heads, q_seq_len // bq, grid_width)
    with jax.named_scope(kernel_name):
        all_out = pl.pallas_call(
            partial(
                flash_attention_kernel,
                mask_value=mask_value,
                grid_width=grid_width,
                bq=bq,
                bkv=bkv,
                bkv_compute=bkv_compute,
                head_dim=head_dim,
                q_layout=q_layout,
                k_layout=k_layout,
                v_layout=v_layout,
                attn_logits_soft_cap=attn_logits_soft_cap,
                mask_function=mask_function,
            ),
            grid_spec=pltpu.PrefetchScalarGridSpec(
                num_scalar_prefetch=num_scalar_prefetch,
                in_specs=in_specs,
                out_specs=out_specs,
                grid=grid,
            ),
            compiler_params=dict(mosaic=mosaic_params),
            out_shape=out_shapes,
            name=kernel_name,
        )(
            fwd_mask_info.data_next,
            fwd_mask_info.block_mask,
            fwd_mask_info.mask_next,
            q if q_layout == QKVLayout.HEAD_DIM_MINOR else q.swapaxes(-1, -2),
            k if k_layout == QKVLayout.HEAD_DIM_MINOR else k.swapaxes(-1, -2),
            v if v_layout == QKVLayout.HEAD_DIM_MINOR else v.swapaxes(-1, -2),
            q_segment_ids,
            kv_segment_ids,
            fwd_mask_info.partial_mask_blocks,
            q_sequence,
        )

    (
        _,
        _,
        _,
        out,
        logsumexp,
    ) = all_out

    if save_residuals:
        assert logsumexp is not None
        logsumexp = logsumexp[..., 0]

    if residual_checkpoint_name is not None:
        out = ad_checkpoint.checkpoint_name(out, name=residual_checkpoint_name)
        if logsumexp is not None:
            logsumexp = ad_checkpoint.checkpoint_name(
                logsumexp, name=residual_checkpoint_name
            )
    if save_residuals:
        return out, (logsumexp,)
    return out


@partial(jax.custom_vjp, nondiff_argnums=(7, 8, 9, 10, 11, 12, 13))
def _splash_attention_custom(
        fwd_mask_info: mask_info_lib.MaskInfo,
        dq_mask_info: Optional[mask_info_lib.MaskInfo],
        dkv_mask_info: Optional[mask_info_lib.MaskInfo],
        q: jax.Array,
        k: jax.Array,
        v: jax.Array,
        segment_ids: Optional[SegmentIds],
        save_residuals: bool,
        mask_value: float,
        is_mqa: bool,
        block_sizes: BlockSizes,
        residual_checkpoint_name: Optional[str],
        mask_function: Optional[MaskFunctionType],
        attn_logits_soft_cap: Optional[float] = None,
) -> SplashCustomReturnType:
    # The forward function does not use the dq and dkv MaskInfos, it just forwards
    # them to the backward function as residuals. This is a way to communicate
    # arbitrary Arrays to the backward function. Since the three MaskInfos are
    # constants there is no overhead in passing them to the backward function as
    # residuals. When sharding computation MaskInfos are partitioned so both the
    # forward and the backward kernels need to work on the relevant slice. If we
    # recomputed the backward MaskInfos in the backward function from the numpy
    # mask then we would not work with the MaskInfo slice relevant to the current
    # device.
    del dq_mask_info, dkv_mask_info

    return _splash_attention_forward(  # pytype: disable=wrong-arg-types
        fwd_mask_info,
        q,
        k,
        v,
        segment_ids,
        mask_value=mask_value,
        is_mqa=is_mqa,
        block_sizes=block_sizes,
        save_residuals=save_residuals,
        attn_logits_soft_cap=attn_logits_soft_cap,
        residual_checkpoint_name=residual_checkpoint_name,
        mask_function=mask_function,
    )


def _splash_attention_fwd(
        fwd_mask_info: mask_info_lib.MaskInfo,
        dq_mask_info: Optional[mask_info_lib.MaskInfo],
        dkv_mask_info: Optional[mask_info_lib.MaskInfo],
        q: jax.Array,
        k: jax.Array,
        v: jax.Array,
        segment_ids: Optional[SegmentIds],
        save_residuals: bool,
        mask_value: float,
        is_mqa: bool,
        block_sizes: BlockSizes,
        residual_checkpoint_name: Optional[str],
        mask_function: Optional[MaskFunctionType],
        attn_logits_soft_cap: Optional[float] = None,
) -> tuple[
    tuple[jax.Array],
    SplashResidualsType,
]:
    if save_residuals:
        raise NotImplementedError("Higher-order AD not supported")

    out, (logsumexp,) = _splash_attention_forward(
        fwd_mask_info,
        q,
        k,
        v,
        segment_ids,
        mask_value=mask_value,
        is_mqa=is_mqa,
        block_sizes=block_sizes,
        save_residuals=True,
        attn_logits_soft_cap=attn_logits_soft_cap,
        residual_checkpoint_name=residual_checkpoint_name,
        mask_function=mask_function,
    )
    return out, (
        q,
        k,
        v,
        segment_ids,
        out,
        logsumexp,
        dq_mask_info,
        dkv_mask_info,
    )


def _flash_attention_dq_kernel(
        # Prefetched inputs
        data_next_ref,
        block_mask_ref,
        mask_next_ref,
        # Inputs
        q_ref,
        k_ref,
        v_ref,
        q_segment_ids_ref,
        kv_segment_ids_ref,
        logsumexp_ref,
        do_ref,
        di_ref,
        mask_ref,
        q_sequence_ref,
        # Outputs
        dq_scratch_ref,
        dq_ref,
        *,
        mask_value: float,
        grid_width: int,
        bq: int,
        bkv: int,
        attn_logits_soft_cap: Optional[float] = None,
        q_layout: QKVLayout,
        k_layout: QKVLayout,
        v_layout: QKVLayout,
        mask_function: Optional[MaskFunctionType],
):
    float32 = jnp.float32
    HEAD_DIM_MINOR = QKVLayout.HEAD_DIM_MINOR

    h, i, j = pl.program_id(0), pl.program_id(1), pl.program_id(2)

    @pl.when(j == 0)
    def init():
        dq_scratch_ref[...] = jnp.zeros_like(dq_scratch_ref)

    _, _, should_run, should_not_mask = _next_nonzero(
        h, i, j, data_next_ref, block_mask_ref, mask_next_ref
    )

    @pl.when(should_run)
    def run():
        q = q_ref[...] if q_layout == HEAD_DIM_MINOR else q_ref[...].T
        # We keep k and v possibly transposed, since they are RHS of dots.
        k = k_ref[...]
        v = v_ref[...]
        logsumexp = jnp.expand_dims(logsumexp_ref[0], -1)
        do = do_ref[...]
        di = jnp.expand_dims(di_ref[0], -1)

        qk_dims = NT_DIM_NUMBERS if k_layout == HEAD_DIM_MINOR else NN_DIM_NUMBERS
        qk_uncapped = lax.dot_general(q, k, qk_dims, preferred_element_type=float32)

        qk = _apply_mask_and_soft_cap(
            qk_uncapped,
            mask_value,
            should_not_mask,
            mask_ref,
            q_sequence_ref,
            q_segment_ids_ref,
            kv_segment_ids_ref,
            attn_logits_soft_cap=attn_logits_soft_cap,
            k_slice=pl.ds(0, bkv),
            k_offset=j * bkv,
            bq=bq,
            mask_function=mask_function,
        )
        p = jnp.exp(qk - logsumexp)
        dp_dims = NT_DIM_NUMBERS if v_layout == HEAD_DIM_MINOR else NN_DIM_NUMBERS
        dp = lax.dot_general(
            do.astype(v.dtype), v, dp_dims, preferred_element_type=jnp.float32,
        )
        ds = (dp - di) * p
        if attn_logits_soft_cap is not None:
            normalized = qk_uncapped / attn_logits_soft_cap
            d = jnp.tanh(normalized)
            g = ds * (1 - d)
            ds = g + g * d

        dq_dims = NN_DIM_NUMBERS if k_layout == HEAD_DIM_MINOR else NT_DIM_NUMBERS
        dq_scratch_ref[...] += lax.dot_general(
            ds.astype(k.dtype), k, dq_dims,
            preferred_element_type=jnp.float32,
        )

    @pl.when(j == grid_width - 1)
    def end():
        dq_ref[...] = dq_scratch_ref[...].astype(dq_ref.dtype)
        dq_scratch_ref[...] = jnp.zeros_like(dq_scratch_ref)


def _splash_attention_bwd_dq(
        q,
        k,
        v,
        segment_ids,
        logsumexp,
        do,
        di,
        *,
        bq: int,
        bkv: int,
        is_mqa: bool,
        mask_info: mask_info_lib.MaskInfo,
        mask_value: float,
        attn_logits_soft_cap: Optional[float],
        q_layout: QKVLayout,
        k_layout: QKVLayout,
        v_layout: QKVLayout,
        mask_function: Optional[MaskFunctionType],
):
    num_q_heads, q_seq_len, head_dim = q.shape
    if is_mqa:
        kv_seq_len = k.shape[0]
        num_kv_heads = 1
    else:
        kv_seq_len = k.shape[1]
        num_kv_heads = k.shape[0]

    if bq > q_seq_len:
        raise ValueError(
            f"{bq=} should not be greater than {q_seq_len=}")
    if bkv > kv_seq_len:
        raise ValueError(
            f"{bkv=} should not be greater than {kv_seq_len=}")

    if not is_mqa and num_q_heads % num_kv_heads != 0:
        raise ValueError(
            f"In MHA, expected number of 'key' heads ({num_kv_heads}) to be a"
            f" multiple of the number of 'query' heads ({num_q_heads})"
        )

    if k.shape != v.shape:
        raise ValueError(
            f"Expected 'key' {k.shape} and 'value' {v.shape} to have the same"
            " shape."
        )

    if bkv % NUM_LANES:
        raise ValueError(f"{bkv=} must be a multiple of {NUM_LANES}.")

    # TODO(amagni/sharadmv): when adding block_compute, make sure that is a
    # multiple of NUM_LANES.

    q_heads_per_kv_head = num_q_heads // num_kv_heads

    if mask_info.data_next is not None:
        grid_width = mask_info.data_next.shape[-1]
    else:
        grid_width = kv_seq_len // bkv

    grid = (num_q_heads, q_seq_len // bq, grid_width)

    def o_index_map(h, i, *_):
        return h, i, 0

    o_spec = pl.BlockSpec(o_index_map, (None, bq, head_dim))

    def q_index_map(h, i, *_):
        return from_head_minor((h, i, 0), q_layout)

    q_spec = pl.BlockSpec(
        q_index_map, from_head_minor((None, bq, head_dim), q_layout)
    )

    def k_index_map(h, i, j, data_next_ref, block_mask_ref, mask_next_ref, *_):
        next_j, *_ = _next_nonzero(
            h, i, j, data_next_ref, block_mask_ref, mask_next_ref
        )
        prefix = () if is_mqa else (_div(h, q_heads_per_kv_head),)
        return from_head_minor((*prefix, next_j, 0), k_layout)

    k_spec = pl.BlockSpec(
        k_index_map,
        from_head_minor(
            (bkv, head_dim) if is_mqa else (None, bkv, head_dim), k_layout),
    )

    def v_index_map(h, i, j, data_next_ref, block_mask_ref, mask_next_ref, *_):
        next_j, *_ = _next_nonzero(
            h, i, j, data_next_ref, block_mask_ref, mask_next_ref
        )
        prefix = () if is_mqa else (_div(h, q_heads_per_kv_head),)
        return from_head_minor((*prefix, next_j, 0), v_layout)

    v_spec = pl.BlockSpec(
        v_index_map,
        from_head_minor(
            (bkv, head_dim) if is_mqa else (None, bkv, head_dim), v_layout),
    )

    def mask_index_map(h, i, j, data_next_ref, block_mask_ref, mask_next_ref, *_):
        _, next_m, *_ = _next_nonzero(
            h, i, j, data_next_ref, block_mask_ref, mask_next_ref
        )
        return next_m, 0, 0

    mask_spec = pl.BlockSpec(mask_index_map, (None, bq, bkv))

    def q_segment_ids_index_map(h, i, j, *_):
        del h, j  # Unused.
        return i, 0

    if segment_ids is not None:

        def kv_segment_ids_index_map(
                h, i, j, data_next_ref, block_mask_ref, mask_next_ref, *_
        ):
            next_j, *_ = _next_nonzero(
                h, i, j, data_next_ref, block_mask_ref, mask_next_ref
            )
            return 0, next_j

        q_segment_spec = pl.BlockSpec(q_segment_ids_index_map, (bq, NUM_LANES))
        kv_segment_spec = pl.BlockSpec(kv_segment_ids_index_map,
                                       (NUM_SUBLANES, bkv))
        q_segment_ids = jax.lax.broadcast_in_dim(
            segment_ids.q, (q_seq_len, NUM_LANES), (0,)
        )
        kv_segment_ids = jax.lax.broadcast_in_dim(
            segment_ids.kv, (NUM_SUBLANES, kv_seq_len), (1,)
        )
    else:
        q_segment_spec = kv_segment_spec = None
        q_segment_ids = kv_segment_ids = None

    do_spec = dq_spec = o_spec

    def logsumexp_index_map(h, i, *_):
        return h, 0, i

    logsumexp = jnp.expand_dims(logsumexp, axis=-2)
    logsumexp_spec = pl.BlockSpec(logsumexp_index_map, (None, 1, bq))
    assert logsumexp.ndim == len(logsumexp_spec.block_shape)

    di = jnp.expand_dims(di, axis=-2)
    di_spec = pl.BlockSpec(logsumexp_index_map, (None, 1, bq))
    assert di.ndim == len(di_spec.block_shape)

    in_specs = [
        q_spec,
        k_spec,
        v_spec,
        q_segment_spec,
        kv_segment_spec,
        logsumexp_spec,
        do_spec,
        di_spec,
    ]
    if mask_info.partial_mask_blocks is not None:
        in_specs.append(mask_spec)
    else:
        in_specs.append(None)

    assert mask_info.partial_mask_blocks is None or mask_info.q_sequence is None

    if mask_info.q_sequence is not None:
        q_sequence = jax.lax.broadcast_in_dim(
            mask_info.q_sequence, (q_seq_len, NUM_LANES), (0,)
        )
        in_specs.append(pl.BlockSpec(q_segment_ids_index_map, (bq, NUM_LANES)))
    else:
        q_sequence = None
        in_specs.append(None)

    out_shapes = [
        jax.ShapeDtypeStruct((bq, head_dim), jnp.float32),
        jax.ShapeDtypeStruct(q.shape, q.dtype),
    ]
    out_specs = [
        pl.BlockSpec(lambda *_: (0, 0), (bq, head_dim)),
        dq_spec,
    ]

    kernel = functools.partial(
        _flash_attention_dq_kernel,
        grid_width=grid_width,
        mask_value=mask_value,
        bq=bq,
        bkv=bkv,
        attn_logits_soft_cap=attn_logits_soft_cap,
        q_layout=q_layout,
        k_layout=k_layout,
        v_layout=v_layout,
        mask_function=mask_function,
    )
    num_scalar_prefetch = 3

    # Attach useful metadata to the custom-call HLO op.
    # Having this information available in an HLO-dump or xprof is valuable for
    # debugging and performance investigation.
    metadata_dict = dict(
        block_q_dq=bq,
        block_kv_dq=bkv,
        q_layout=q_layout,
        k_layout=k_layout,
        v_layout=v_layout,
        is_mqa=is_mqa,
        mask_value=mask_value,
        is_segmented=segment_ids is not None,
        attn_logits_soft_cap=attn_logits_soft_cap,
    )

    mosaic_params = pltpu.encode_kernel_regeneration_metadata(metadata_dict)
    mosaic_params.update(
        dimension_semantics=("arbitrary", "arbitrary", "arbitrary"),
        flags={"XLA_TPU_FORCE_LP_LLO_SCHEDULER": True},
    )

    kernel_name = get_kernel_name(is_mqa, False, segment_ids is not None, "dq")
    with jax.named_scope(kernel_name):
        _, dq = pl.pallas_call(
            kernel,
            grid_spec=pltpu.PrefetchScalarGridSpec(
                num_scalar_prefetch=num_scalar_prefetch,
                in_specs=in_specs,
                out_specs=out_specs,
                grid=grid,
            ),
            out_shape=out_shapes,
            compiler_params=dict(mosaic=mosaic_params),
            name=kernel_name,
        )(
            mask_info.data_next,
            mask_info.block_mask,
            mask_info.mask_next,
            q if q_layout == QKVLayout.HEAD_DIM_MINOR else q.swapaxes(-1, -2),
            k if k_layout == QKVLayout.HEAD_DIM_MINOR else k.swapaxes(-1, -2),
            v if v_layout == QKVLayout.HEAD_DIM_MINOR else v.swapaxes(-1, -2),
            q_segment_ids,
            kv_segment_ids,
            logsumexp,
            do,
            di,
            mask_info.partial_mask_blocks,
            q_sequence,
        )
    return dq


def _flash_attention_dkv_kernel(
        # Prefetched inputs
        data_next_ref,
        block_mask_ref,
        mask_next_ref,
        # Inputs
        q_ref,
        k_ref,
        v_ref,
        q_segment_ids_ref,
        kv_segment_ids_ref,
        logsumexp_ref,
        do_ref,
        di_ref,
        mask_ref,
        q_sequence_ref,
        # Outputs
        dq_scratch_ref,
        dk_scratch_ref,
        dv_scratch_ref,
        dq_ref,
        dk_ref,
        dv_ref,
        *,
        num_q_heads: int,
        num_kv_heads: int,
        mask_value: float,
        grid_width: int,
        bq: int,
        bkv_compute: int,
        is_mqa: bool,
        attn_logits_soft_cap: Optional[float],
        q_layout: QKVLayout,
        k_layout: QKVLayout,
        v_layout: QKVLayout,
        bkv: int,
        mask_function: Optional[MaskFunctionType],
):
    HEAD_DIM_MINOR = QKVLayout.HEAD_DIM_MINOR
    kv_index, q_head_index, q_index = (
        pl.program_id(0),
        pl.program_id(1),
        pl.program_id(2),
    )
    should_initialize = q_index == 0

    q_heads_per_kv_heads = None
    q_head_index_per_kv_head = None

    # Consider this situation:
    # Q_heads:   0, 1, 2, 3, 4, 5, 6, 7
    # KV_heads:  0,    1,    2,    3
    # The gradient scratch buffers should be initialized for Q_heads 0, 2, 4, 6
    # (first Q_heads to 'see' a new KV_head).
    # The gradient output buffers should be written for Q_heads 1, 3, 5, 7 (last
    # Q_heads to 'see' the current KV_head).

    # We can use the same logic for both MQA and GA (grouped attention).
    # But for MQA there is no need for the rem instruction, so we skip it.
    if is_mqa:
        should_initialize = jnp.logical_and(should_initialize, q_head_index == 0)
    elif num_kv_heads < num_q_heads:
        q_heads_per_kv_heads = num_q_heads // num_kv_heads
        q_head_index_per_kv_head = lax.rem(q_head_index, q_heads_per_kv_heads)
        should_initialize = jnp.logical_and(
            should_initialize, q_head_index_per_kv_head == 0
        )

    @pl.when(should_initialize)
    def init():
        dk_scratch_ref[...] = jnp.zeros_like(dk_scratch_ref)
        dv_scratch_ref[...] = jnp.zeros_like(dv_scratch_ref)

    _, _, should_run, should_not_mask = _next_nonzero(
        q_head_index,
        q_index,
        kv_index,
        data_next_ref,
        block_mask_ref,
        mask_next_ref,
        next_i=True,
    )

    def body(i, _):

        slice_k = pl.ds(i * bkv_compute, bkv_compute)
        q = q_ref[...]  # We keep q potentially transposed, since it's always RHS

        def _load_kv(ref, layout):
            if layout == HEAD_DIM_MINOR:
                return pl.load(ref, (slice_k, slice(None)))
            return pl.load(ref, (slice(None), slice_k)).T

        k = _load_kv(k_ref, k_layout)
        v = _load_kv(v_ref, v_layout)
        logsumexp = pl.load(logsumexp_ref, (pl.ds(1), slice(None)))
        do = do_ref[...]
        di = pl.load(di_ref, (pl.ds(1), slice(None)))

        qk_dims = NT_DIM_NUMBERS if q_layout == HEAD_DIM_MINOR else NN_DIM_NUMBERS
        qk_uncapped = lax.dot_general(
            k, q, qk_dims, preferred_element_type=jnp.float32
        )

        qk = _apply_mask_and_soft_cap(
            qk_uncapped,
            mask_value,
            should_not_mask,
            mask_ref,
            q_sequence_ref,
            q_segment_ids_ref,
            kv_segment_ids_ref,
            attn_logits_soft_cap=attn_logits_soft_cap,
            k_slice=slice_k,
            k_offset=kv_index * bkv + i * bkv_compute,
            bq=bq,
            k_in_lanes=False,
            mask_function=mask_function,
        )
        p = jnp.exp(qk - logsumexp)
        dv = lax.dot(p.astype(do.dtype), do, preferred_element_type=jnp.float32)
        dv = dv.astype(dv_scratch_ref.dtype) + pl.load(
            dv_scratch_ref, (slice_k, slice(None))
        )
        pl.store(dv_scratch_ref, (slice_k, slice(None)), dv)

        dp = lax.dot_general(
            v, do, NT_DIM_NUMBERS,
            preferred_element_type=jnp.float32,
        )
        ds = (dp - di) * p
        if attn_logits_soft_cap is not None:
            normalized = qk_uncapped / attn_logits_soft_cap
            d = jnp.tanh(normalized)
            g = ds * (1 - d)
            ds = g + g * d
        dk_dims = NN_DIM_NUMBERS if q_layout == HEAD_DIM_MINOR else NT_DIM_NUMBERS
        dk = lax.dot_general(
            ds.astype(do.dtype), q, dk_dims, preferred_element_type=jnp.float32
        )
        dk = dk.astype(dk_scratch_ref.dtype) + pl.load(
            dk_scratch_ref, (slice_k, slice(None))
        )
        pl.store(dk_scratch_ref, (slice_k, slice(None)), dk)
        if dq_scratch_ref is not None or dq_ref is not None:
            dq = lax.dot_general(
                ds.T.astype(k.dtype), k, NN_DIM_NUMBERS,
                preferred_element_type=jnp.float32,
            )
            if dq_scratch_ref is not None:
                # Compute block size != memory block size
                dq_scratch_ref[...] += dq
            else:
                # Compute block size == memory block size
                assert dq_ref is not None
                dq_ref[...] = dq.astype(dq_ref.dtype)

    if dq_scratch_ref is not None:
        dq_scratch_ref[...] = jnp.zeros_like(dq_scratch_ref)
    elif dq_scratch_ref is None and dq_ref is not None:
        dq_ref[...] = jnp.zeros_like(dq_ref)

    @pl.when(should_run)
    def run():
        num_iters = (
                k_ref.shape[0 if k_layout is HEAD_DIM_MINOR else 1] // bkv_compute
        )
        lax.fori_loop(0, num_iters, body, None, unroll=True)

    if dq_scratch_ref is not None:
        assert dq_ref is not None
        dq_ref[...] = dq_scratch_ref[...].astype(dq_ref.dtype)

    should_write = q_index == grid_width - 1
    if is_mqa:
        should_write = jnp.logical_and(
            should_write, q_head_index == num_q_heads - 1
        )
    elif num_kv_heads < num_q_heads:
        should_write = jnp.logical_and(
            should_write, q_head_index_per_kv_head == q_heads_per_kv_heads - 1
        )

    @pl.when(should_write)
    def end():
        dk_ref[...] = dk_scratch_ref[...].astype(dk_ref.dtype)
        dv_ref[...] = dv_scratch_ref[...].astype(dv_ref.dtype)
        if dq_scratch_ref is not None:
            dq_scratch_ref[...] = jnp.zeros_like(dq_scratch_ref)

        dk_scratch_ref[...] = jnp.zeros_like(dk_scratch_ref)
        dv_scratch_ref[...] = jnp.zeros_like(dv_scratch_ref)


def _splash_attention_bwd_dkv(
        q,
        k,
        v,
        segment_ids,
        logsumexp,
        do,
        di,
        *,
        bq: int,
        bkv: int,
        bkv_compute: int,
        is_mqa: bool,
        mask_info: mask_info_lib.MaskInfo,
        mask_value: float,
        attn_logits_soft_cap: Optional[float],
        use_fused_bwd_kernel: bool,
        q_layout: QKVLayout,
        k_layout: QKVLayout,
        v_layout: QKVLayout,
        mask_function: Optional[MaskFunctionType],
):
    num_q_heads, q_seq_len, head_dim = q.shape
    if is_mqa:
        num_kv_heads, kv_seq_len = 1, k.shape[0]
    else:
        num_kv_heads, kv_seq_len, _ = k.shape

    if bq > q_seq_len:
        raise ValueError(
            f"{bq=} should not be greater than {q_seq_len=}")
    if bkv > kv_seq_len:
        raise ValueError(
            f"{bkv=} should not be greater than {kv_seq_len=}")
    if bkv_compute > bkv:
        raise ValueError(
            f"{bkv_compute=} should not be greater than {bkv=}")
    if bkv % bkv_compute:
        raise ValueError(
            f"{bkv=} should be a multiple of {bkv_compute=}")

    if not is_mqa and num_q_heads % num_kv_heads != 0:
        raise ValueError(
            f"In MHA, expected number of 'key' heads ({num_kv_heads}) to be a"
            f" multiple of the number of 'query' heads ({num_q_heads})"
        )

    if k.shape != v.shape:
        raise ValueError(
            f"Expected 'key' {k.shape} and 'value' {v.shape} to have the same"
            " shape."
        )

    q_heads_per_kv_head = num_q_heads // num_kv_heads

    if mask_info.data_next is not None:
        grid_width = mask_info.data_next.shape[-2]
    else:
        grid_width = q_seq_len // bq

    grid = (
        kv_seq_len // bkv,
        num_q_heads,
        grid_width,
    )

    def o_index_map(
            kv_index,
            head_index,
            q_index,
            data_next_ref,
            block_mask_ref,
            mask_next_ref=None,
    ):
        next_i, *_ = _next_nonzero(
            head_index,
            q_index,
            kv_index,
            data_next_ref,
            block_mask_ref,
            mask_next_ref,
            next_i=True,
        )
        return head_index, next_i, 0

    o_spec = pl.BlockSpec(o_index_map, (None, bq, head_dim))

    def q_index_map(
            kv_index,
            head_index,
            q_index,
            data_next_ref,
            block_mask_ref,
            mask_next_ref=None,
    ):
        next_i, *_ = _next_nonzero(
            head_index,
            q_index,
            kv_index,
            data_next_ref,
            block_mask_ref,
            mask_next_ref,
            next_i=True,
        )
        return from_head_minor((head_index, next_i, 0), q_layout)

    q_spec = pl.BlockSpec(
        q_index_map, from_head_minor((None, bq, head_dim), q_layout))

    def k_index_map(kv_index, head_index, *_):
        prefix = () if is_mqa else (_div(head_index, q_heads_per_kv_head),)
        return from_head_minor((*prefix, kv_index, 0), k_layout)

    k_spec = pl.BlockSpec(
        k_index_map,
        from_head_minor(
            (bkv, head_dim) if is_mqa else (None, bkv, head_dim),
            k_layout,
        ),
    )

    def v_index_map(kv_index, head_index, *_):
        prefix = () if is_mqa else (_div(head_index, q_heads_per_kv_head),)
        return from_head_minor((*prefix, kv_index, 0), v_layout)

    v_spec = pl.BlockSpec(
        v_index_map,
        from_head_minor(
            (bkv, head_dim) if is_mqa else (None, bkv, head_dim),
            v_layout,
        ),
    )

    if use_fused_bwd_kernel:
        def dq_index_map(kv_index, head_index, q_index, *_):
            return (kv_index, head_index, q_index, 0)

        dq_spec = pl.BlockSpec(dq_index_map, (None, None, bq, head_dim))
        dq_shape = jax.ShapeDtypeStruct((kv_seq_len // bkv, *q.shape), q.dtype)
        if bkv == bkv_compute:
            dq_scratch_spec = dq_scratch_shape = None
        else:
            dq_scratch_spec = pl.BlockSpec(lambda *_: (0, 0), (bq, head_dim))
            dq_scratch_shape = jax.ShapeDtypeStruct((bq, head_dim), jnp.float32)
    else:
        dq_spec = dq_shape = dq_scratch_spec = dq_scratch_shape = None

    def dkv_index_map(kv_index, head_index, *_):
        prefix = () if is_mqa else (_div(head_index, q_heads_per_kv_head),)
        return (*prefix, kv_index, 0)

    dk_spec = dv_spec = pl.BlockSpec(
        dkv_index_map,
        (bkv, head_dim) if is_mqa else (None, bkv, head_dim),
    )

    def mask_index_map(
            kv_index,
            head_index,
            q_index,
            data_next_ref,
            block_mask_ref,
            mask_next_ref,
    ):
        _, next_m, *_ = _next_nonzero(
            head_index,
            q_index,
            kv_index,
            data_next_ref,
            block_mask_ref,
            mask_next_ref,
            next_i=True,
        )
        return next_m, 0, 0

    mask_spec = pl.BlockSpec(mask_index_map, (None, bkv, bq))

    def q_segment_ids_index_map(
            kv_index,
            head_index,
            q_index,
            data_next_ref,
            block_mask_ref,
            mask_next_ref=None,
    ):
        next_i, *_ = _next_nonzero(
            head_index,
            q_index,
            kv_index,
            data_next_ref,
            block_mask_ref,
            mask_next_ref,
            next_i=True,
        )
        return 0, next_i

    if segment_ids is not None:
        def kv_segment_ids_index_map(kv_index, *_):
            return kv_index, 0

        q_segment_spec = pl.BlockSpec(q_segment_ids_index_map, (NUM_SUBLANES, bq))
        kv_segment_spec = pl.BlockSpec(kv_segment_ids_index_map,
                                       (bkv, NUM_LANES))
        q_segment_ids = jax.lax.broadcast_in_dim(
            segment_ids.q, (NUM_SUBLANES, q_seq_len), (1,)
        )
        kv_segment_ids = jax.lax.broadcast_in_dim(
            segment_ids.kv, (kv_seq_len, NUM_LANES), (0,)
        )
    else:
        q_segment_spec = kv_segment_spec = None
        q_segment_ids = kv_segment_ids = None

    do_spec = o_spec

    def logsumexp_index_map(
            kv_index,
            head_index,
            q_index,
            data_next_ref,
            block_mask_ref,
            mask_next_ref=None,
    ):
        next_i, *_ = _next_nonzero(
            head_index,
            q_index,
            kv_index,
            data_next_ref,
            block_mask_ref,
            mask_next_ref,
            next_i=True,
        )
        return head_index, 0, next_i

    assert logsumexp.shape == di.shape == (num_q_heads, q_seq_len)
    # TODO(apaszke): Remove the sublane expansion once Mosaic has all retilings
    logsumexp_shape = (num_q_heads, NUM_SUBLANES, q_seq_len)
    logsumexp = jnp.broadcast_to(jnp.expand_dims(logsumexp, -2), logsumexp_shape)
    logsumexp_spec = pl.BlockSpec(logsumexp_index_map, (None, NUM_SUBLANES, bq))
    assert logsumexp.ndim == len(logsumexp_spec.block_shape)

    # TODO(apaszke): Remove the sublane expansion once Mosaic has all retilings
    di = jnp.broadcast_to(jnp.expand_dims(di, -2), logsumexp_shape)
    di_spec = pl.BlockSpec(logsumexp_index_map, (None, NUM_SUBLANES, bq))
    assert di.ndim == len(di_spec.block_shape)

    in_specs = [
        q_spec,
        k_spec,
        v_spec,
        q_segment_spec,
        kv_segment_spec,
        logsumexp_spec,
        do_spec,
        di_spec,
    ]
    if mask_info.partial_mask_blocks is not None:
        in_specs.append(mask_spec)
    else:
        in_specs.append(None)

    if mask_info.q_sequence is not None:
        in_specs.append(pl.BlockSpec(q_segment_ids_index_map, (NUM_SUBLANES, bq)))
        q_sequence = jax.lax.broadcast_in_dim(
            mask_info.q_sequence, (NUM_SUBLANES, q_seq_len), (1,)
        )
    else:
        q_sequence = None
        in_specs.append(None)

    out_shapes = [
        dq_scratch_shape,
        jax.ShapeDtypeStruct((bkv, head_dim), jnp.float32),
        jax.ShapeDtypeStruct((bkv, head_dim), jnp.float32),
        dq_shape,
        jax.ShapeDtypeStruct(k.shape, k.dtype),
        jax.ShapeDtypeStruct(v.shape, v.dtype),
    ]
    out_specs = [
        dq_scratch_spec,
        pl.BlockSpec(lambda *_: (0, 0), (bkv, head_dim)),
        pl.BlockSpec(lambda *_: (0, 0), (bkv, head_dim)),
        dq_spec,
        dk_spec,
        dv_spec,
    ]

    kernel = functools.partial(
        _flash_attention_dkv_kernel,
        mask_value=mask_value,
        num_q_heads=num_q_heads,
        num_kv_heads=num_kv_heads,
        is_mqa=is_mqa,
        grid_width=grid_width,
        bq=bq,
        bkv_compute=bkv_compute,
        attn_logits_soft_cap=attn_logits_soft_cap,
        q_layout=q_layout,
        k_layout=k_layout,
        v_layout=v_layout,
        bkv=bkv,
        mask_function=mask_function,
    )
    num_scalar_prefetch = 3

    # Attach useful metadata to the custom-call HLO op.
    # Having this information available in an HLO-dump or xprof is valuable for
    # debugging and performance investigation.
    metadata_dict = dict(
        block_q_dkv=bq,
        block_kv_dkv=bkv,
        block_kv_dkv_compute=bkv_compute,
        q_layout=q_layout,
        k_layout=k_layout,
        v_layout=v_layout,
        use_fused_bwd_kernel=use_fused_bwd_kernel,
        is_mqa=is_mqa,
        mask_value=mask_value,
        is_segmented=segment_ids is not None,
        attn_logits_soft_cap=attn_logits_soft_cap,
    )

    mosaic_params = pltpu.encode_kernel_regeneration_metadata(metadata_dict)
    # We set all dimensions to arbitrary because:
    # 1) for kv_seq_len, the splash attention prefetch schedule assumes no
    #    megacore
    # 2) for heads, we are reducing over heads
    # 3) for q_seq_len, we are reducing over it to compute dkv
    mosaic_params.update(
        dimension_semantics=("arbitrary", "arbitrary", "arbitrary"),
        flags={"XLA_TPU_FORCE_LP_LLO_SCHEDULER": True},
    )

    kernel_name = get_kernel_name(is_mqa, False, segment_ids is not None, "dkv")
    with jax.named_scope(kernel_name):
        _, _, _, dq_unreduced, dk, dv = pl.pallas_call(
            kernel,
            grid_spec=pltpu.PrefetchScalarGridSpec(
                num_scalar_prefetch=num_scalar_prefetch,
                in_specs=in_specs,
                out_specs=out_specs,
                grid=grid,
            ),
            out_shape=out_shapes,
            compiler_params=dict(mosaic=mosaic_params),
            name=kernel_name,
        )(
            mask_info.data_next,
            mask_info.block_mask,
            mask_info.mask_next,
            q if q_layout == QKVLayout.HEAD_DIM_MINOR else q.swapaxes(-1, -2),
            k if k_layout == QKVLayout.HEAD_DIM_MINOR else k.swapaxes(-1, -2),
            v if v_layout == QKVLayout.HEAD_DIM_MINOR else v.swapaxes(-1, -2),
            q_segment_ids,
            kv_segment_ids,
            logsumexp,
            do,
            di,
            mask_info.partial_mask_blocks,
            q_sequence,
        )
    if use_fused_bwd_kernel:
        assert dq_unreduced is not None
        dq = dq_unreduced.sum(axis=0)
    else:
        assert dq_unreduced is None
        dq = None
    return dq, dk, dv


def _splash_attention_bwd(
        save_residuals: bool,
        mask_value: float,
        is_mqa: bool,
        block_sizes: BlockSizes,
        residual_checkpoint_name: Optional[str],
        mask_function: Optional[MaskFunctionType],
        attn_logits_soft_cap: Optional[float],
        res: SplashResidualsType,
        do: jax.Array,
) -> tuple[
    Optional[mask_info_lib.MaskInfo],  # fwd_mask_info
    Optional[mask_info_lib.MaskInfo],  # dq_mask_info
    Optional[mask_info_lib.MaskInfo],  # dvk_mask_info
    jax.Array,  # q
    jax.Array,  # k
    jax.Array,  # v
    Optional[SegmentIds],  # segmend_ids
]:
    del save_residuals, residual_checkpoint_name
    if not block_sizes.has_backward_blocks:
        raise ValueError("Need to specify backward blocks.")
    bq_dq, bkv_dq = block_sizes.block_q_dq, block_sizes.block_kv_dq
    bq_dkv, bkv_dkv_memory, bkv_dkv_compute = (
        block_sizes.block_q_dkv,
        block_sizes.block_kv_dkv,
        block_sizes.block_kv_dkv_compute,
    )
    use_fused_bwd_kernel = block_sizes.use_fused_bwd_kernel
    (
        q,
        k,
        v,
        segment_ids,
        o,
        logsumexp,
        dq_mask_info,
        dkv_mask_info,
    ) = res

    # di: [num_heads, q_seq_len]
    di = jnp.einsum("hsd,hsd->hs", o.astype(jnp.float32), do.astype(jnp.float32))  # pytype: disable=attribute-error
    dq, dk, dv = _splash_attention_bwd_dkv(
        q,
        k,
        v,
        segment_ids,
        logsumexp,
        do,
        di,
        bq=bq_dkv,
        bkv=bkv_dkv_memory,
        bkv_compute=bkv_dkv_compute,
        is_mqa=is_mqa,
        mask_info=dkv_mask_info,
        mask_value=mask_value,
        attn_logits_soft_cap=attn_logits_soft_cap,
        use_fused_bwd_kernel=use_fused_bwd_kernel,
        q_layout=block_sizes.q_layout,
        k_layout=block_sizes.k_layout,
        v_layout=block_sizes.v_layout,
        mask_function=mask_function,
    )
    if not use_fused_bwd_kernel:
        assert dq is None
        dq = _splash_attention_bwd_dq(
            q,
            k,
            v,
            segment_ids,
            logsumexp,
            do,
            di,
            bq=bq_dq,
            bkv=bkv_dq,
            is_mqa=is_mqa,
            mask_info=dq_mask_info,
            mask_value=mask_value,
            attn_logits_soft_cap=attn_logits_soft_cap,
            q_layout=block_sizes.q_layout,
            k_layout=block_sizes.k_layout,
            v_layout=block_sizes.v_layout,
            mask_function=mask_function,
        )
    # Match the signature of the fwd function.
    assert dq is not None
    return (
        None,  # fwd_mask_info
        None,  # dq_mask_info
        None,  # dvk_mak_info
        dq,  # q
        dk,  # k
        dv,  # v
        None,  # segment_ids
    )


_splash_attention_custom.defvjp(_splash_attention_fwd, _splash_attention_bwd)


@partial(
    jax.jit,
    static_argnames=[
        "is_mqa",
        "block_sizes",
        "save_residuals",
        "mask_value",
        "attn_logits_soft_cap",
        "residual_checkpoint_name",
        "mask_function",
    ],
)
def _splash_attention(
        fwd_mask_info: mask_info_lib.MaskInfo,
        dq_mask_info: Optional[mask_info_lib.MaskInfo],
        dkv_mask_info: Optional[mask_info_lib.MaskInfo],
        q: jax.Array,
        k: jax.Array,
        v: jax.Array,
        segment_ids: Optional[SegmentIds] = None,
        *,
        is_mqa: bool,
        block_sizes: Optional[BlockSizes],
        save_residuals: bool,
        mask_value: float,
        attn_logits_soft_cap: Optional[float],
        residual_checkpoint_name: Optional[str],
        mask_function: Optional[MaskFunctionType],
) -> SplashCustomReturnType:
    return _splash_attention_custom(
        fwd_mask_info,
        dq_mask_info,
        dkv_mask_info,
        q,
        k,
        v,
        segment_ids,
        mask_value=mask_value,
        is_mqa=is_mqa,
        block_sizes=block_sizes,
        save_residuals=save_residuals,
        attn_logits_soft_cap=attn_logits_soft_cap,
        residual_checkpoint_name=residual_checkpoint_name,
        mask_function=mask_function,
    )


@jax.tree_util.register_pytree_node_class
class SplashAttentionKernel:

    def __init__(
            self,
            fwd_mask_info: mask_info_lib.MaskInfo,
            dq_mask_info: Optional[mask_info_lib.MaskInfo],
            dkv_mask_info: Optional[mask_info_lib.MaskInfo],
            **kwargs,
    ):
        self.kwargs = kwargs
        self.fwd_mask_info = fwd_mask_info
        self.dq_mask_info = dq_mask_info
        self.dkv_mask_info = dkv_mask_info

    def __call__(self, *args, **kwargs) -> SplashCustomReturnType:
        return _splash_attention(
            self.fwd_mask_info,
            self.dq_mask_info,
            self.dkv_mask_info,
            *args,
            **kwargs,
            **self.kwargs,
        )

    def manual_sharding_spec(self, sharding: jax.sharding.NamedSharding):
        """Returns a value that can be used as a shard_map partition spec for the kernel."""
        if self.fwd_mask_info.data_next is not None:
            block_mask_shape = self.fwd_mask_info.data_next.shape
            try:
                shard_shape = sharding.shard_shape(block_mask_shape)
            except ValueError as exc:
                raise ValueError(
                    "The sharding must divide the mask blocks evenly between devices"
                ) from exc
            if block_mask_shape[-1] != shard_shape[-1]:
                raise ValueError("Sharding the kv sequence dimension is not supported")
        spec = sharding.spec
        assert len(spec) == 2
        replicated = jax.sharding.PartitionSpec()
        # Shard q_sequence over the sequence dimension only.
        q_sequence_spec = jax.sharding.PartitionSpec(spec[1])
        mask_info_specs = mask_info_lib.MaskInfo(  # pytype: disable=wrong-arg-types
            data_next=spec if self.fwd_mask_info.data_next is not None else None,
            mask_next=spec if self.fwd_mask_info.mask_next is not None else None,
            block_mask=spec if self.fwd_mask_info.block_mask is not None else None,
            partial_mask_blocks=replicated
            if self.fwd_mask_info.partial_mask_blocks is not None
            else None,
            q_sequence=q_sequence_spec
            if self.fwd_mask_info.q_sequence is not None
            else None,
        )
        return SplashAttentionKernel(
            mask_info_specs,
            mask_info_specs if self.dq_mask_info is not None else None,
            mask_info_specs if self.dkv_mask_info is not None else None,
            **self.kwargs,
        )

    def tree_flatten(self):
        return (
            (self.fwd_mask_info, self.dq_mask_info, self.dkv_mask_info),
            self.kwargs,
        )

    @classmethod
    def tree_unflatten(cls, kwargs, values):
        fwd_mask_info, dq_mask_info, dkv_mask_info = values
        # NamedTuples are not preserved during pytree serialization.
        dq_mask_info = (
            mask_info_lib.MaskInfo(*dq_mask_info)
            if dq_mask_info is not None
            else None
        )
        dkv_mask_info = (
            mask_info_lib.MaskInfo(*dkv_mask_info)
            if dkv_mask_info is not None
            else None
        )
        return SplashAttentionKernel(
            mask_info_lib.MaskInfo(*fwd_mask_info),
            dq_mask_info,
            dkv_mask_info,
            **kwargs,
        )


def _make_splash_attention(
        mask: Union[np.ndarray, mask_lib.MultiHeadMask],
        *,
        block_sizes: Optional[BlockSizes] = None,
        is_mqa: bool,
        save_residuals: bool = False,
        mask_value: float = DEFAULT_MASK_VALUE,
        attn_logits_soft_cap: Optional[float] = None,
        downcast_smem_data: bool = True,
        head_shards: int,
        q_seq_shards: int,
        residual_checkpoint_name: Optional[str] = None,
):
    if len(mask.shape) != 3:
        raise ValueError(f'Unexpected mask shape: {mask.shape}')

    if isinstance(mask, np.ndarray):
        mask = mask_lib.MultiHeadMask(
            [mask_lib.NumpyMask(head_mask) for head_mask in mask]
        )

    if block_sizes is None:
        block_sizes = BlockSizes.get_default()
    fwd_mask_info, mask_function_fwd = mask_info_lib.process_mask(
        mask,
        (block_sizes.block_q, block_sizes.block_kv),
        downcast_smem_data=downcast_smem_data,
        head_shards=head_shards,
        q_seq_shards=q_seq_shards,
    )

    fwd_mask_info = tree_util.tree_map(jnp.array, fwd_mask_info)

    dq_mask_info = None
    dkv_mask_info = None
    if block_sizes.has_backward_blocks:
        if block_sizes.use_fused_bwd_kernel:
            dq_mask_info = None
        else:
            bq_dq, bkv_dq = block_sizes.block_q_dq, block_sizes.block_kv_dq
            dq_mask_info, mask_function_dq = mask_info_lib.process_mask(
                mask,
                (bq_dq, bkv_dq),
                downcast_smem_data=downcast_smem_data,
                head_shards=head_shards,
                q_seq_shards=q_seq_shards,
            )
            assert (mask_function_fwd is None) == (mask_function_dq is None)
            dq_mask_info = tree_util.tree_map(jnp.array, dq_mask_info)
        bq_dkv, bkv_dkv = block_sizes.block_q_dkv, block_sizes.block_kv_dkv
        dkv_mask_info, mask_function_dkv = mask_info_lib.process_mask_dkv(
            mask,
            (bq_dkv, bkv_dkv),
            downcast_smem_data=downcast_smem_data,
            head_shards=head_shards,
            q_seq_shards=q_seq_shards,
            shrink_grid=not block_sizes.use_fused_bwd_kernel,
        )
        assert (mask_function_fwd is None) == (mask_function_dkv is None)

        dkv_mask_info = tree_util.tree_map(jnp.array, dkv_mask_info)

    return SplashAttentionKernel(
        fwd_mask_info,
        dq_mask_info,
        dkv_mask_info,
        block_sizes=block_sizes,
        is_mqa=is_mqa,
        save_residuals=save_residuals,
        mask_value=mask_value,
        attn_logits_soft_cap=attn_logits_soft_cap,
        residual_checkpoint_name=residual_checkpoint_name,
        mask_function=mask_function_fwd,
    )


make_splash_mha = partial(_make_splash_attention, is_mqa=False)
make_splash_mqa = partial(_make_splash_attention, is_mqa=True)

make_splash_mha_single_device = partial(
    make_splash_mha, is_mqa=False, head_shards=1, q_seq_shards=1
)

make_splash_mqa_single_device = partial(
    make_splash_mha, is_mqa=True, head_shards=1, q_seq_shards=1
)



from abc import abstractmethod, ABCMeta
from contextlib import contextmanager
from contextvars import ContextVar
from dataclasses import (
    dataclass,
    field,
    fields,
    is_dataclass
)
from typing import ClassVar, Union
import warnings

import numpy as np
from jax.api_util import flatten_fun
import jax.interpreters.partial_eval as pe
from jax.core import Shape
from jax.tree_util import register_pytree_with_keys_class

from jax.typing import DTypeLike

import jax.numpy as jnp
from jax.dtypes import float0
import optax

from typing import Any, Optional, Tuple

from plum import (
    dispatch,
    parametric,
    Dispatcher,
    Function
)
from abc import ABC
from itertools import count

from functools import partial, wraps
from itertools import chain

import jax
from jax.api_util import flatten_fun_nokwargs
from jax import core
import jax.extend.linear_util as lu
from jax import tree_util


class ArrayValue(metaclass=ABCMeta):
    """Helper class that provides a standard way to create an ABC using
    inheritance.
    """
    __slots__ = ()
    shape = None
    e_num_val = None
    is_registered_by_pJit = False


ArrayValue.register(jax.Array)

_dispatch = Dispatcher()

_primitive_ids = count()

COMMUTATIVE_OPS = frozenset([
    "add",
    "bitwise_and",
    "bitwise_or",
    "bitwise_xor",
    "eq",
    "max",
    "min",
    "mul",
    "ne",
])

ELEMENTWISE_UNOPS = frozenset([
    "abs",
    "acos",
    "acosh",
    "asin",
    "asinh",
    "atan",
    "atanh",
    "bessel_i0e",
    "bessel_i1e",
    "cbrt",
    "ceil",
    "clz",
    "conj",
    "convert_element_type",
    "copy",
    "cos",
    "cosh",
    "digamma",
    "erf_inv",
    "erf",
    "erfc",
    "exp",
    "expm1",
    "floor",
    "imag",
    "integer_pow",
    "is_finite",
    "lgamma",
    "log1p",
    "log",
    "logistic",
    "neg",
    "not",
    "population_count",
    "real",
    "reduce_precision",
    "round",
    "rsqrt",
    "sign",
    "sin",
    "sinh",
    "sqrt",
    "tan",
    "tanh",
])

ELEMENTWISE_BINOPS = frozenset([
    "add",
    "and",
    "atan2",
    "complex",
    "div",
    "eq",
    "ge",
    "gt",
    "igamma_grad_a",
    "igamma",
    "igammac",
    "le",
    "lt",
    "max",
    "min",
    "mul",
    "ne",
    "nextafter",
    "or",
    "pow",
    "random_gamma_grad",
    "rem",
    "shift_left",
    "shift_right_arithmetic",
    "shift_right_logical",
    "sub",
    "xor",
])

REDUCTION_OPS = frozenset([
    "argmax",
    "argmin",
    "reduce_and",
    "reduce_max",
    "reduce_min",
    "reduce_or",
    "reduce_prod",
    "reduce_sum",
    "reduce_xor",
])

CUMULATIVE_REDUCTION_OPS = frozenset([
    "cumlogsumexp",
    "cummax",
    "cummin",
    "cumprod",
    "cumsum",
])

_GENERAL = -2
_SPECIALIZED = -1


def _with_implicit_flat(fun: lu.WrappedFun) -> lu.WrappedFun:
    return _implicit_outer(_implicit_inner(fun))


@lu.transformation
def _implicit_outer(*in_vals):
    with core.new_main(ImplicitArrayTrace) as main:
        outs = yield (main, *in_vals), {}
        del main
    yield outs


@lu.transformation
def _implicit_inner(main, *in_vals):
    trace = main.with_cur_sublevel()
    in_tracers = [
        ImplicitArrayTracer(trace, val) if isinstance(val, ImplicitArray) else val
        for val in in_vals
    ]
    outs = yield in_tracers, {}
    out_vals = [trace.full_raise(t).value for t in outs]
    yield out_vals


def use_implicit_args(f):
    """
    Decorator which allows a function to accept arguments which subclass ImplicitArray, possibly
    including further ImplicitArray instances as children.
    Any number of arguments (including 0) may be ImplicitArrays.
    """

    @wraps(f)
    def implicit_f(*args, **kwargs):
        flat_args, in_tree = tree_flatten_with_implicit((args, kwargs))
        f_flat, out_tree = flatten_fun(lu.wrap_init(f), in_tree)
        f_wrapped = _with_implicit_flat(f_flat)
        outs_flat = f_wrapped.call_wrapped(*flat_args)
        return out_tree().unflatten(outs_flat)

    return implicit_f


def aux_field(metadata: Optional[Union[dict, Any]] = None, **kwargs):
    metadata = dict(metadata) if metadata else {}
    metadata["implicit_array_aux"] = True
    return field(metadata=metadata, **kwargs)


class UninitializedAval(Exception):
    def __init__(self, kind):
        super().__init__(_AVAL_ERROR_MESSAGE.format(kind))


class _AvalDescriptor:
    def __set_name__(self, owner, name):
        self._name = f"_{name}"

    def __get__(self, obj, owner=None):
        if obj is None:
            return None
        result = getattr(obj, self._name, None)
        if result is None:
            raise UninitializedAval(kind=self._name[1:])
        return result

    def __set__(self, obj, value):
        setattr(obj, self._name, value)


_aval_discovery = ContextVar("aval_discovery", default=False)


@contextmanager
def _aval_discovery_context():
    token = _aval_discovery.set(True)
    try:
        yield
    finally:
        _aval_discovery.reset(token)


@dataclass
class _ImplicitArrayBase(ArrayValue, ABC):
    commute_ops: ClassVar[bool] = True
    default_shape: ClassVar[Optional[Shape]] = None
    default_dtype: ClassVar[Optional[DTypeLike]] = None

    shape: Optional[Shape] = aux_field(kw_only=True, default=None)
    dtype: DTypeLike = aux_field(kw_only=True, default=None)


@dataclass
class ImplicitArray(_ImplicitArrayBase):
    """
    Abstract class for representing an abstract array of a given shape/dtype without actually instantiating it.
    Subclasses must implement the materialize method, which defines the relationship between the implicit array
    and the value it represents. Subclasses are valid arguments to functions decorated with qax.use_implicit_args.

    All subclasses are automatically registered as pytrees using jax.tree_util.register_pytree_with_keys_class.
    Any dataclass attributes added will be included as children, unless they are decorated with qax.aux_field
    in which case they are passed as auxiliary data during flattening.

    The represented shape and dtype may be defined in any of the following ways:
        - Explicitly passing shape/dtype keyword arguments at initialization
        - Overriding the default_shape/default_dtype class variables
        - Overriding the compute_shape/compute_dtype methods, which are called during __post_init__
        - Overriding __post_init__ and manually setting shape/dtype before calling super().__post_init__
        - None of the above, in which case an shape/dtype will be inferred by by running jax.eval_shape()
          on the subclass"s materialize method.
    """

    shape = _AvalDescriptor()
    dtype = _AvalDescriptor()

    def __post_init__(self):
        try:
            aval = _get_materialization_aval(self)
        except UninitializedAval:
            # Materialization depends on currently uninitialized shape/dtype
            aval = None

        shape = None
        try:
            shape = self.shape
        except UninitializedAval as e:
            shape = self.shape = self.compute_shape()

        if aval is not None:
            if shape is None:
                self.shape = aval.shape
            elif shape != aval.shape:
                warnings.warn(f"ImplicitArray shape {shape} does not match materialization shape {aval.shape}")
        elif shape is None:
            raise UninitializedAval("shape")

        dtype = None
        try:
            dtype = self.dtype
        except UninitializedAval as e:
            dtype = self.dtype = self.compute_dtype()

        if dtype is None and aval is None:
            # We have a shape but not a dtype, try once again to infer the dtype
            aval = _get_materialization_aval(self)

        if aval is not None:
            if dtype is None:
                self.dtype = aval.dtype
            elif dtype != aval.dtype:
                warnings.warn(f"ImplicitArray dtype {dtype} does not match materialization dtype {aval.dtype}")
        elif dtype is None:
            raise UninitializedAval("dtype")

    def compute_shape(self):
        """
        Override this method if the subclass instance"s shape should be computed based on its other properties.
        Returns: shape
        """
        return self.default_shape

    def compute_dtype(self):
        """
        Override this method if the subclass instance"s dtype should be computed based on its other properties.
        Returns: dtype
        """
        return self.default_dtype

    @property
    def aval(self):
        return core.ShapedArray(self.shape, self.dtype)

    @classmethod
    def default_handler(cls, primitive, *args, params=None):
        if params is None:
            params = {}
        return materialize_handler(primitive, *args, params=params)

    @abstractmethod
    def materialize(self):
        pass

    def tree_flatten_with_keys(self):
        children = []
        aux_data = []
        for name, is_aux in _get_names_and_aux(self):
            try:
                value = getattr(self, name)
            except UninitializedAval:
                if not _aval_discovery.get():
                    raise
                value = None
            if is_aux:
                aux_data.append(value)
            else:
                children.append((name, value))

        return children, aux_data

    @classmethod
    def tree_unflatten(cls, aux_data, children):
        child_it = iter(children)
        aux_it = iter(aux_data)
        obj = cls.__new__(cls)
        for name, is_aux in _get_names_and_aux(cls):
            value = next(aux_it if is_aux else child_it)
            setattr(obj, name, value)

        return obj

    def handle_primitive(self, primitive, *args, params):
        handler = lu.wrap_init(partial(get_primitive_handler(primitive), primitive))
        use_params = params

        if len(args) == 2 and self.commute_ops:
            args, use_params = _maybe_swap_args(primitive.name, args, use_params)

        # maybe_kwargs = {"params": params} if params else {}
        flat_args, in_tree = flatten_one_implicit_layer((args, params))
        flat_handler, out_tree = flatten_fun(handler, in_tree)

        result = use_implicit_args(flat_handler.call_wrapped)(*flat_args)
        return jax.tree_util.tree_unflatten(out_tree(), result)

    def __init_subclass__(cls, commute_ops=True, **kwargs):
        super().__init_subclass__(**kwargs)

        if not is_dataclass(cls):
            raise TypeError(f"{cls.__name__} must be a dataclass")
        core.pytype_aval_mappings[cls] = lambda x: x.aval
        register_pytree_with_keys_class(cls)
        return cls


def _get_names_and_aux(obj):
    for val in fields(obj):
        yield val.name, bool(val.metadata.get("implicit_array_aux"))


def _materialize_all(it):
    return [materialize_nested(val) if isinstance(val, ImplicitArray) else val for val in it]


def _maybe_swap_args(op_name, args, params):
    if isinstance(args[0], ImplicitArray):
        return args, params
    if op_name in COMMUTATIVE_OPS:
        return args[::-1], params

    return args, params


class ImplicitArrayTracer(core.Tracer):
    def __init__(self, trace, value):
        super().__init__(trace)
        self.value = value

    @property
    def aval(self):
        if isinstance(self.value, ImplicitArray):
            return self.value.aval
        return core.get_aval(self.value)

    def full_lower(self):
        if isinstance(self.value, ImplicitArray):
            return self

        return core.full_lower(self.value)


class ImplicitArrayTrace(core.Trace):
    pure = lift = lambda self, val: ImplicitArrayTracer(self, val)

    def process_primitive(self, primitive, tracers, params):
        outs = NotImplemented
        vals = [t.value for t in tracers]
        implicit_idx = next(i for i, v in enumerate(vals) if isinstance(v, ImplicitArray))

        # First try to handle the primitive using custom handlers
        outs = vals[implicit_idx].handle_primitive(primitive, *vals, params=params)

        if outs is NotImplemented:
            # For higher order primitives most users won"t implement custom
            # logic, so there shouldn"t be a warning
            if primitive.name in _default_handlers:
                outs = _default_handlers[primitive.name](primitive, *vals, params=params)
            else:
                warnings.warn(
                    f"Primitive {primitive.name} was not handled by class {vals[implicit_idx].__class__.__name__}, so implicit args will be materialized.")

        if outs is NotImplemented:
            outs = vals[implicit_idx].default_handler(primitive, *vals, params=params)

        if primitive.multiple_results:
            return [ImplicitArrayTracer(self, out) for out in outs]
        return ImplicitArrayTracer(self, outs)


def wrap_jaxpr(jaxpr, vals_with_implicits, return_closed=True):
    if isinstance(jaxpr, jax.core.ClosedJaxpr):
        literals = jaxpr.literals
        jaxpr = jaxpr.jaxpr
    else:
        literals = []

    wrapped_fn = lu.wrap_init(use_implicit_args(partial(core.eval_jaxpr, jaxpr)))
    flat_args, in_tree = jax.tree_util.tree_flatten((literals, *vals_with_implicits))
    flat_fn, out_tree = flatten_fun_nokwargs(wrapped_fn, in_tree)

    new_jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(flat_fn, [core.get_aval(v) for v in flat_args])

    ret = (jax.core.ClosedJaxpr(new_jaxpr, consts),) if return_closed else (new_jaxpr, consts)
    return *ret, flat_args, out_tree()


def _transform_jaxpr_output(jaxpr, jaxpr_args, orig_out_struct, out_transform):
    def eval_fn(literals, *args):
        output = use_implicit_args(partial(core.eval_jaxpr, jaxpr.jaxpr))(literals, *args)
        unflattened_output = orig_out_struct.unflatten(output)
        return out_transform(unflattened_output)

    wrapped = lu.wrap_init(eval_fn)

    flat_args, in_tree = jax.tree_util.tree_flatten((jaxpr.literals, *jaxpr_args))
    flat_fn, out_tree = flatten_fun_nokwargs(wrapped, in_tree)
    new_jaxpr, _, consts = pe.trace_to_jaxpr_dynamic(flat_fn, [core.get_aval(v) for v in flat_args])

    return jax.core.ClosedJaxpr(new_jaxpr, consts), out_tree()


def _match_branches(branches, arg_vals):
    out_avals = []
    new_jaxprs = []
    flat_inputs = None
    branch_out_struct = None
    for branch in branches:
        new_jaxpr, flat_inputs, branch_out_struct = wrap_jaxpr(branch, arg_vals)
        new_jaxprs.append((new_jaxpr, branch_out_struct))
        out_avals.append(
            branch_out_struct.unflatten(
                jax.eval_shape(
                    partial(core.eval_jaxpr, new_jaxpr.jaxpr), new_jaxpr.literals, *flat_inputs
                )
            )
        )

    out_transforms = get_common_prefix_transforms(out_avals)
    new_branches = []
    out_struct = None
    for (new_jaxpr, orig_out_struct), transform in zip(new_jaxprs, out_transforms):
        new_jaxpr, out_struct = _transform_jaxpr_output(new_jaxpr, flat_inputs, orig_out_struct, transform)
        new_branches.append(new_jaxpr)

    return tuple(new_branches), out_struct, flat_inputs


def _handle_cond(primitive, *vals, params):
    cond_val, *arg_vals = vals
    subfuns, bind_params = primitive.get_bind_params(params)

    new_branches, out_struct, flat_inputs = _match_branches(params["branches"], arg_vals)
    bind_params["branches"] = new_branches
    bind_params["linear"] = _broadcast_tuple(params["linear"], arg_vals)

    outs = primitive.bind(*subfuns, cond_val, *flat_inputs, **bind_params)
    return jax.tree_util.tree_unflatten(out_struct, outs)


def _handle_remat2(primitive, *vals, params):
    subfuns, bind_params = primitive.get_bind_params(params)
    new_jaxpr, consts, flat_inputs, out_tree = wrap_jaxpr(bind_params["jaxpr"], vals, return_closed=False)
    new_jaxpr = pe.convert_constvars_jaxpr(new_jaxpr)
    bind_params["jaxpr"] = new_jaxpr
    outs = primitive.bind(*subfuns, *consts, *flat_inputs, **bind_params)
    return jax.tree_util.tree_unflatten(out_tree, outs)


def _handle_pjit(primitive, *vals, params):
    new_jaxpr, flat_inputs, out_tree = wrap_jaxpr(params["jaxpr"], vals)
    donated_invars = _broadcast_tuple(params["donated_invars"], vals)
    in_shardings = _broadcast_tuple(params["in_shardings"], vals)
    out_shardings = _broadcast_tuple(params["out_shardings"], out_tree)

    subfuns, bind_params = primitive.get_bind_params(params)
    bind_params["jaxpr"] = new_jaxpr
    bind_params["donated_invars"] = donated_invars
    bind_params["in_shardings"] = in_shardings
    bind_params["out_shardings"] = out_shardings
    outs = primitive.bind(*subfuns, *flat_inputs, **bind_params)
    return jax.tree_util.tree_unflatten(out_tree, outs)


_default_handlers = {
    "cond": _handle_cond,
    "remat2": _handle_remat2,
    "pjit": _handle_pjit,
}


def materialize_handler(primitive, *vals, params):
    vals = _materialize_all(vals)
    subfuns, bind_params = primitive.get_bind_params(params)
    result = use_implicit_args(primitive.bind)(*subfuns, *vals, **bind_params)
    return result


def _broadcast_tuple(t, trees):
    if isinstance(trees, jax.tree_util.PyTreeDef):
        trees = jax.tree_util.tree_unflatten(trees, range(trees.num_leaves))
    assert len(t) == len(trees)
    return tuple(chain.from_iterable(
        (tuple_val for _ in jax.tree_util.tree_leaves(tree))
        for tuple_val, tree in zip(t, trees)
    ))


def _get_materialization_aval(imp_arr):
    with _aval_discovery_context(), _filter_materialization_warnings():
        result = jax.eval_shape(
            partial(materialize_nested, full=True),
            imp_arr
        )
    return result


@contextmanager
def _filter_materialization_warnings():
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Primitive.*was not handled")
        yield


_AVAL_ERROR_MESSAGE = (
    "{} was not set during initialization. Shape and dtype may be set by:"
    "\n\t1. Directly passing them as keyword arguments to ImplicitArray instances"
    "\n\t2. Overriding the default_shape/default_dtype class attributes"
    "\n\t3. Overriding the compute_shape/compute_dtype methods"
    "\n\t4. Overriding __post_init__ and setting their values there"
    "\n\t5. None of the above, in which case `materialize()` will be called in an attempt to infer them."
    " If their values are required in order to compute the materialization this will be unsuccessful."
)


class _EmptyNodeCls:
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
        return cls._instance


EmptyNode = _EmptyNodeCls()

tree_util.register_pytree_node(
    _EmptyNodeCls,
    lambda node: ((), None),
    lambda _, __: EmptyNode
)


def combine_leaf_predicate(base_fn, is_leaf):
    @wraps(base_fn)
    def new_fn(*args, new_is_leaf=None):
        if new_is_leaf is None:
            combined_is_leaf = is_leaf
        else:
            def combined_is_leaf(arg):
                return is_leaf(arg) or new_is_leaf(arg)
        return base_fn(*args, is_leaf=combined_is_leaf)

    return new_fn


def leaf_predicate(x):
    return isinstance(x, (ImplicitArray, _EmptyNodeCls))


tree_map_with_implicit = combine_leaf_predicate(jax.tree_map, leaf_predicate)
tree_map_with_path_with_implicit = combine_leaf_predicate(tree_util.tree_map_with_path, leaf_predicate)
tree_flatten_with_implicit = combine_leaf_predicate(tree_util.tree_flatten, leaf_predicate)
tree_flatten_with_path_with_implicit = combine_leaf_predicate(tree_util.tree_flatten_with_path, leaf_predicate)
tree_leaves_with_implicit = combine_leaf_predicate(tree_util.tree_leaves, leaf_predicate)
tree_structure_with_implicit = combine_leaf_predicate(tree_util.tree_structure, leaf_predicate)


def flatten_one_implicit_layer(tree):
    def is_leaf_below_node(node, x):
        return isinstance(x, ImplicitArray) and x is not node

    def replace_subtree_implicits(node):
        return tree_util.tree_map(lambda _: 1, node, is_leaf=partial(is_leaf_below_node, node))

    prototype = tree_map_with_implicit(replace_subtree_implicits, tree)
    struct = tree_util.tree_structure(prototype)

    leaves = tree_leaves_with_implicit(tree)
    leaves = list(chain.from_iterable(
        tree_util.tree_leaves(leaf, is_leaf=partial(is_leaf_below_node, leaf))
        if isinstance(leaf, ImplicitArray) else
        [leaf] for leaf in leaves
    ))
    return leaves, struct


def implicit_depth(tree):
    leaves = tree_leaves_with_implicit(tree)
    depth = 0
    while True:
        next_leaves = []
        any_implicit = False
        for leaf in leaves:
            if not isinstance(leaf, ImplicitArray):
                continue
            any_implicit = True
            next_leaves.extend(flatten_one_implicit_layer(leaf)[0])

        if not any_implicit:
            return depth

        depth += 1
        leaves = next_leaves


def _map_leaves_with_implicit_path(f, leaves, is_leaf, path_prefix=()):
    mapped_leaves = []
    for idx, leaf in enumerate(leaves):
        path = path_prefix + (idx,)
        if not isinstance(leaf, ImplicitArray) or is_leaf(path, leaf):
            mapped_leaves.append(f(leaf))
            continue

        subtree, substruct = flatten_one_implicit_layer(leaf)
        mapped_subtree = _map_leaves_with_implicit_path(
            f,
            subtree,
            is_leaf=is_leaf,
            path_prefix=path
        )
        mapped_leaves.append(tree_util.tree_unflatten(substruct, mapped_subtree))
    return mapped_leaves


def _get_pruning_transform(tree, materialization_paths):
    if not materialization_paths:
        return lambda x: x

    def is_leaf(path, leaf):
        return path in materialization_paths

    def materialize_subtrees(tree):
        leaves, struct = tree_flatten_with_implicit(tree)

        mapped_leaves = _map_leaves_with_implicit_path(partial(materialize_nested, full=True), leaves, is_leaf)
        return tree_util.tree_unflatten(struct, mapped_leaves)

    return materialize_subtrees


def get_common_prefix_transforms(trees):
    """
    Given an iterable of pytrees which have the same structure after all
    ImplicitArray instances are materialized, return a list of callables
    which will transform each tree into the largest common structure
    obtainable via materialization of ImplicitArrays.
    """
    if len(trees) <= 1:
        return [lambda x: x for _ in trees]

    all_leaves, structures = zip(*(tree_flatten_with_implicit(tree) for tree in trees))
    post_materialization_avals = [core.get_aval(leaf) for leaf in all_leaves[0]]
    for i, (leaves, structure) in enumerate(zip(all_leaves[1:], structures[1:]), 1):
        if structure != structures[0]:
            raise ValueError('Trees do not have the same structure after materialization')

        for leaf, expected_aval in zip(leaves, post_materialization_avals):
            aval = core.get_aval(leaf)
            if not (aval.shape == expected_aval.shape and aval.dtype == expected_aval.dtype):
                raise ValueError(
                    f'Trees do not have the same avals after materialization. Tree 0: {expected_aval}, Tree {i}: {aval}'
                )

    # Stack will contain tuples of (path, nodes)
    # path = a sequence of integers specifying which child
    # was taken at each _flatten_one_implicit_layer call
    # or the first flatten_with_implicit call
    # nodes = one node from each tree
    stack = []

    all_leaves = []
    for tree in trees:
        all_leaves.append(tree_leaves_with_implicit(tree))

    for i, nodes in enumerate(zip(*all_leaves)):
        stack.append(((i,), nodes))

    materialization_paths = set()
    while stack:
        path_prefix, nodes = stack.pop()
        if not any(isinstance(node, ImplicitArray) for node in nodes):
            continue

        all_leaves, all_structures = zip(*(
            flatten_one_implicit_layer(node) for node in nodes
        ))
        node_structures = set(all_structures)
        if len(node_structures) > 1:
            materialization_paths.add(path_prefix)
            continue

        aval_diff = False
        for leaves in zip(*all_leaves):
            first_aval = core.get_aval(leaves[0])
            shape = first_aval.shape
            dtype = first_aval.dtype
            for leaf in leaves[1:]:
                aval = core.get_aval(leaf)
                if not (aval.shape == shape and aval.dtype == dtype):
                    materialization_paths.add(path_prefix)
                    aval_diff = True
            if aval_diff:
                break

        if aval_diff:
            continue

        for i, leaf_group in enumerate(zip(*all_leaves)):
            stack.append((path_prefix + (i,), leaf_group))

    return [_get_pruning_transform(tree, materialization_paths) for tree in trees]


def materialize_nested(implicit_arr, full=False):
    """
    Materialize an ImplicitArray instance, handling the case where implicit_arr.materialize()
    involves further ImplicitArray instances.
    Arguments:
        implicit_arr: An ImplicitArray instance
        full: If True, repeatedly materialize until the result is a concrete array
    Returns:
        The materialized array
    """
    while isinstance(implicit_arr, ImplicitArray):
        wrapped = lu.wrap_init(type(implicit_arr).materialize)
        flat, in_tree = flatten_one_implicit_layer((implicit_arr,))
        flat_fn, out_tree = flatten_fun_nokwargs(wrapped, in_tree)
        out_flat = use_implicit_args(flat_fn.call_wrapped)(*flat)
        implicit_arr = jax.tree_util.tree_unflatten(out_tree(), out_flat)

        if not full:
            break

    return implicit_arr


def get_lax_primitive_by_name(name):
    return getattr(jax.lax, f"{name}_p")


def get_primitive_handler(primitive):
    if isinstance(primitive, str):
        primitive = get_lax_primitive_by_name(primitive)
    handler = _dispatch.functions.get(primitive)
    if handler is None:
        def _not_impl_handler(primitive: jax.core.Primitive, *args, **kwargs):
            return NotImplemented

        _not_impl_handler.__doc__ = "Default handler for {primitive.name}"
        handler = Function(_not_impl_handler)
        handler.register(_not_impl_handler, precedence=-1e9)
        handler.__name__ = f"{primitive.name}_{next(_primitive_ids)}"
        _dispatch.functions[primitive] = handler
    return handler


def primitive_handler(primitives, precedence=0):
    if isinstance(primitives, (str, jax.core.Primitive)):
        primitives = [primitives]

    def decorator(fn):
        for primitive in primitives:
            handler = get_primitive_handler(primitive)
            handler.register(fn, precedence=precedence)

    return decorator


def default_handler(primitive, *args, **params):
    subfuns, bind_params = primitive.get_bind_params(params)
    return primitive.bind(*subfuns, *args, **bind_params)


class _ComplementMeta(type):
    def __instancecheck__(self, x):
        a, b = self.type_parameter
        return (
                a is None or (
                isinstance(x, a) and not isinstance(x, b)
        )
        )


@parametric
class Complement(metaclass=_ComplementMeta):
    """
    Relative complement
    I.e. Complement[A, B] = A - B
    """

    @classmethod
    @dispatch
    def __init_type_parameter__(
            cls,
            a: Optional[Any],
            b: Optional[Any],
    ):
        return a, b

    @classmethod
    @dispatch
    def __le_type_parameter__(
            cls,
            left: Tuple[Optional[Any], Optional[Any]],
            right: Tuple[Optional[Any], Optional[Any]],
    ):
        a_left, b_left = left
        a_right, b_right = right

        return issubclass(a_left, a_right) and issubclass(b_right, b_left)


def vmap_all_but_one(f, axis, out_ndim=0):
    """
    Repeatedly calls vmap to map over all axes except for `axis.`
    All args will be mapped on the same dimensions.
    """

    @wraps(f)
    def inner(*args):
        n_dim = args[0].ndim
        if axis >= n_dim:
            raise ValueError(f'Axis {axis} is out of bounds for array of dimension {n_dim}')
        fn = f
        vmap_dim = 1
        out_dim = out_ndim
        for i in reversed(range(n_dim)):
            if i == axis:
                vmap_dim = 0
                out_dim = 0
            else:
                fn = jax.vmap(fn, vmap_dim, out_dim)
        return fn(*args)

    return inner


def freeze_subtrees(optimizer: optax.GradientTransformation, label_fn, use_scalar_zeros=False):
    """
    Utility which wraps an optimizer such that subtrees specified by
    label_fn will receive zeros as updates.
    Subtrees to be frozen should be labeled with "freeze"
    and all other subtrees should be labeled with "train"
    """
    multi_transformed_optimizer = optax.multi_transform(
        {
            'freeze': set_to_zero_scalar() if use_scalar_zeros else optax.set_to_zero(),
            'train': optimizer
        },
        label_fn
    )

    def new_update(grads, opt_state, params):
        def map_float0(param, grad):
            if grad.dtype == float0:
                return jnp.zeros((), param.dtype) if use_scalar_zeros else jnp.zeros_like(param)
            return grad

        fixed_grads = jax.tree_map(map_float0, params, grads)
        return multi_transformed_optimizer.update(fixed_grads, opt_state, params)

    return optax.GradientTransformation(
        multi_transformed_optimizer.init,
        new_update
    )


def freeze_keys(optimizer: optax.GradientTransformation, arr_type, keys,
                use_scalar_zeros=False) -> optax.GradientTransformation:
    keys = set(keys)

    def label_leaf(leaf):
        if not isinstance(leaf, arr_type):
            return 'train'

        children, aux_data = leaf.tree_flatten_with_keys()
        labels = ['freeze' if key in keys else 'train' for key, _ in children]
        struct = leaf.tree_unflatten(aux_data, labels)
        return struct

    def label_fn(root):
        return jax.tree_map(label_leaf, root, is_leaf=lambda x: isinstance(x, arr_type))

    return freeze_subtrees(optimizer, label_fn, use_scalar_zeros=use_scalar_zeros)


def apply_updates(params: optax.Params, updates: optax.Updates) -> optax.Params:
    """
    Like optax.apply_updates, but updates can be SymbolicConstant instances
    """
    updates_flat, update_struct = tree_util.tree_flatten(updates, is_leaf=lambda x: isinstance(x, SymbolicConstant))
    semi_flat_params = update_struct.flatten_up_to(params)

    updated_flat = use_implicit_args(optax.apply_updates)(semi_flat_params, updates_flat)
    updated = update_struct.unflatten(updated_flat)
    return updated


def set_to_zero_scalar() -> optax.GradientTransformation:
    """
    Returns a gradient transformation that sets all gradients to 0 in order to
    make downstream constant folding cheaper.
    """

    def init_fn(params):
        del params
        return optax.EmptyState()

    def update_fn(updates, state, params=None):
        return jax.tree_map(lambda x: jnp.zeros((), x.dtype), updates), state

    return optax.GradientTransformation(init_fn, update_fn)


# WARNING: This file is obviously super incomplete, and is
# currently just for convenience in testing.


def _get_shape_dtype(x, shape, dtype):
    if shape is None:
        shape = np.shape(x)
    else:
        shape = jax.core.canonicalize_shape(shape)

    if dtype is None:
        jax.lax.dtype(x)
    return shape, dtype


def _out_shape_dtype(primitive, *args, **kwargs):
    out_aval = jax.eval_shape(
        partial(default_handler, primitive, **kwargs),
        *(jax.core.get_aval(x) for x in args)
    )
    return jax.tree_map(
        lambda x: (x.shape, x.dtype),
        out_aval
    )


def symbolic_zero_like(x, shape=None, dtype=None):
    dtype = jax.lax.dtype(x) if dtype is None else dtype
    return symbolic_full_like(x, 0, shape=shape, dtype=dtype)


def symbolic_full_like(x, fill_value, shape=None, dtype=None):
    shape, _ = _get_shape_dtype(x, shape, None)
    if dtype is None:
        dtype = jax.lax.dtype(fill_value)

    return SymbolicConstant(fill_value, shape=shape, dtype=dtype)


@dataclass
class SymbolicConstant(ImplicitArray):
    value: Any = aux_field()
    weak_type: bool = aux_field(default=False)

    # commute_ops: ClassVar[bool] = True
    # default_shape: ClassVar[Optional[Shape]] = None
    # default_dtype: ClassVar[Optional[DTypeLike]] = None
    # shape: Optional[Shape] = aux_field(kw_only=True, default=None)
    # dtype: DTypeLike = aux_field(kw_only=True, default=None)
    def __post_init__(self):
        super().__post_init__()
        with jax.ensure_compile_time_eval():
            self.value = jnp.asarray(self.value, dtype=self.dtype)

    def compute_dtype(self):
        return jax.lax.dtype(self.value)

    def materialize(self):
        return jnp.full(self.shape, self.value, dtype=self.dtype)

    def copy(self):
        return jax.tree_map(lambda x: x, self)


@use_implicit_args
def broadcast_to(val, shape):
    return jnp.broadcast_to(val, shape)


@use_implicit_args
def astype(val, dtype):
    return val.astype(dtype)


@primitive_handler([
    "reshape",
    "broadcast_in_dim",
    "reduce_min",
    "reduce_max",
    "reduce_or",
    "reduce_and"
])
def unchanged_value_op(primitive, sym: SymbolicConstant, **kwargs):
    out_shape, out_dtype = _out_shape_dtype(primitive, sym, **kwargs)
    return SymbolicConstant(sym.value, shape=out_shape, dtype=out_dtype)


def _op_and_reshape(primitive, lhs, rhs, flip=False):
    """
    Close over one arg so we can do math at tracing time, but let the other one get traced
    """
    if flip:
        lhs, rhs = (rhs, lhs)

    @use_implicit_args
    def inner(arg):
        other = rhs
        if flip:
            arg, other = (other, arg)

        result = default_handler(primitive, arg, other)
        return result

    return inner(rhs)


def special_case_binop(name, identity=None, annihilator=None, flip=False):
    lhs_type = SymbolicConstant
    rhs_type = Complement[ArrayValue, SymbolicConstant]
    if flip:
        lhs_type, rhs_type = rhs_type, lhs_type

    @primitive_handler(name, precedence=_SPECIALIZED)
    def handler(primitive, lhs: lhs_type, rhs: rhs_type, **kwargs):
        out_shape, out_dtype = _out_shape_dtype(primitive, lhs, rhs, **kwargs)
        with jax.ensure_compile_time_eval():
            if lhs.value == identity:
                return broadcast_to(astype(rhs, out_dtype), out_shape)

            if lhs.value == annihilator:
                return SymbolicConstant(lhs.value, shape=out_shape, dtype=out_dtype)

            return _op_and_reshape(primitive, lhs.value, rhs)


special_case_binop("add", identity=0)
special_case_binop("mul", identity=1, annihilator=0)
special_case_binop("and", annihilator=0)
special_case_binop("or", identity=0)
special_case_binop("xor", identity=0)

special_case_binop("sub", identity=0, flip=True)
special_case_binop("div", identity=1, flip=True)
special_case_binop("exp", identity=1, flip=True)

special_case_binop("min", identity=float("inf"), annihilator=float("-inf"))
special_case_binop("max", identity=float("-inf"), annihilator=float("inf"))


def eval_default_handler(primitive, *args, **kwargs):
    with jax.ensure_compile_time_eval():
        result = primitive.bind(*args, **kwargs)
    return result


@primitive_handler(ELEMENTWISE_UNOPS, precedence=_GENERAL)
def handle_unop(primitive, sym: SymbolicConstant, **kwargs):
    print(f"Handling {primitive} with {sym}")
    new_val = eval_default_handler(primitive, sym.value, **kwargs)
    return symbolic_full_like(sym, new_val)


@primitive_handler(ELEMENTWISE_BINOPS, precedence=_GENERAL)
def handle_binop(primitive, lhs: SymbolicConstant, rhs: SymbolicConstant, **kwargs):
    out_shape, out_dtype = _out_shape_dtype(primitive, lhs, rhs, **kwargs)
    new_val = eval_default_handler(primitive, lhs.value, rhs.value, **kwargs)
    return symbolic_full_like(lhs, new_val, shape=out_shape, dtype=out_dtype)


@primitive_handler(["reduce_sum", "reduce_prod"])
def reduce_sum(primitive, sym: SymbolicConstant, *, axes):
    out_shape, out_dtype = _out_shape_dtype(primitive, sym, axes=axes)
    with jax.ensure_compile_time_eval():
        if sym.value == 0:
            return SymbolicConstant(0, shape=out_shape, dtype=out_dtype)

        orig_size = np.prod(sym.shape)
        new_size = np.prod(out_shape)

        n_combined = orig_size // new_size

        new_val = sym.value
        if primitive.name == "reduce_sum":
            new_val = new_val * n_combined
        else:
            new_val = new_val ** n_combined

    return SymbolicConstant(new_val, shape=out_shape, dtype=out_dtype)


@primitive_handler("select_n")
def handle_select_n(primitive, cond_val, *arg_vals: SymbolicConstant):
    if len(set(val.value.item() for val in arg_vals)) != 1:
        return NotImplemented

    out_shape, out_dtype = _out_shape_dtype(primitive, cond_val, *arg_vals)
    return SymbolicConstant(arg_vals[0].value, shape=out_shape, dtype=out_dtype)
